import logging
from colorsys import hls_to_rgb
from typing import TYPE_CHECKING, Any

from vfbLib.typing import GuidePropertiesDict, LinkDict
from vfbLib.ufo.vfb2ufo import vfb2ufo_label_codes

if TYPE_CHECKING:
    from vfbLib.typing import (
        AnchorDict,
        MMGuidesDict,
        MMHintsDict,
        MMNode,
    )
    from vfbLib.ufo.builder import VfbToUfoBuilder
    from vfbLib.ufo.tth import TTGlyphHints


logger = logging.getLogger(__name__)


class VfbToUfoGlyph:
    def __init__(self, builder: "VfbToUfoBuilder | None" = None) -> None:
        self.builder = builder
        self.anchors: "list[AnchorDict]" = []
        self.guide_properties = GuidePropertiesDict(h=[], v=[])
        self.hintmasks: list[tuple[str, int]] = []
        self.labels: dict[str, int] = {}
        self.lib: dict[str, Any] = {}
        self.links: LinkDict = LinkDict(x=[], y=[])
        self.mm_anchors: list[Any] | None = None
        self.mm_components: list[Any] = []
        self.mm_guides: "MMGuidesDict | None" = None
        self.mm_hints: "MMHintsDict" = {"h": [], "v": []}
        self.mm_metrics: list[tuple[int, int]] = []
        self.mm_nodes: "list[MMNode]" = []
        self.mm_mask_nodes: "list[MMNode]" = []
        self.mm_mask_metrics: list[tuple[int, int]] = []
        self.name: str | None = None
        self.note: str | None = None
        self.point_labels: dict[int, str] = {}
        self.ps_hinting_options = dict[str, int | list[int]]
        self.rename_points: dict[str, str]
        self.tt_glyph_hints: "TTGlyphHints | None" = None
        self.tth_commands: list[dict[str, str | bool]] = []
        self.unicodes: list[int] = []

    def get_point_label(self, index: int, code: str, start_count: int = 1) -> str:
        if self.mm_components:
            # Composite: We must add the label to the referenced glyph
            if self.builder is None:
                logger.error(
                    "To compile composite TrueType hinting, you must supply the"
                    "VfbToUfoBuilder to VfbToUfoGlyph.__init__()"
                )
                raise ValueError

            # Find the right component the point index belongs to
            orig_index = index
            total_num_nodes = 0
            for i, mm_component in enumerate(self.mm_components):
                component_name = self.builder.glyphOrder[mm_component["gid"]]
                component = self.builder.glyph_masters[component_name]
                num_nodes = len(component.mm_nodes)
                if index < num_nodes:
                    # Add the component index to the label to make it unique enough.
                    return f"{component.get_point_label(index, code)}-{i}"

                index -= num_nodes
                total_num_nodes += num_nodes

            # Side bearings may be hinted
            if orig_index == total_num_nodes:
                return "lsb"
            elif orig_index == total_num_nodes + 1:
                return "rsb"
            logger.error(
                f"Could not find point {orig_index} for hinted composite '{self.name}'."
                " TrueType hinting will be broken in UFO glyph."
            )
            return "invalid"

        if index in self.point_labels:
            # We already have a label for this point index
            return self.point_labels[index]

        # Special points
        num_nodes = len(self.mm_nodes)
        if index == num_nodes:
            return "lsb"
        elif index == num_nodes + 1:
            return "rsb"

        # Make a new label
        label_short = vfb2ufo_label_codes[code]
        i = start_count
        label = "%s%02d" % (label_short, i)
        while label in self.labels:
            i += 1
            label = "%s%02d" % (label_short, i)
        self.labels[label] = index
        self.point_labels[index] = label
        return label

    def set_mark(self, hue) -> None:
        self.lib["public.markColor"] = "%0.4f,%0.4f,%0.4f,1" % hls_to_rgb(
            h=hue / 255, l=0.8, s=0.76
        )

    def set_mask(self, data) -> None:
        self.mm_mask_nodes = data.get("nodes", [])

    def __eq__(self, other) -> bool:
        if len(self.mm_components) == len(other.mm_components):
            if self.name == other.name:
                return True

        return False

    def __gt__(self, other) -> bool:
        ns = len(self.mm_components)
        no = len(other.mm_components)
        if ns > no:
            return True
        if ns == no:
            if self.name > other.name:
                return True

        return False

    def __lt__(self, other) -> bool:
        ns = len(self.mm_components)
        no = len(other.mm_components)
        if ns < no:
            return True
        if ns == no:
            if self.name < other.name:
                return True

        return False


class IndexVfbToUfoGlyph(VfbToUfoGlyph):
    """
    A glyph that does not convert point indices to point labels, but returns the
    original point index again.
    """

    def get_point_label(self, index: int, code: str) -> int:
        return index



from vfbLib.parsers.base import BaseParser
from vfbLib.typing import PrimaryInstanceDict


class AnisotropicInterpolationsParser(BaseParser):
    def _parse(self) -> list[list[tuple[int, int]]]:
        # The graph used for anisotropic interpolation maps, for all axes.
        values = []
        while True:
            axis_values = []
            try:
                n = self.read_value()
                axis_values = [(self.read_value(), self.read_value()) for _ in range(n)]
                values.append(axis_values)
            except EOFError:
                return values


class AxisMappingsCountParser(BaseParser):
    def _parse(self) -> list[int]:
        # An array of axis mapping counts per axis, for all 4 axes whether they exist
        # or not.
        # 0300 0000  0000 0000  0000 0000  0000 0000
        # -> [3, 0, 0, 0]
        return [self.read_uint32() for _ in range(4)]


class AxisMappingsParser(BaseParser):
    def _parse(self) -> list[tuple[float, float]]:
        # 10 pairs of (user, design) coordinates per axis.
        # Look at "Axis Mappings Count" to find out which mappings are used in each
        # axis.
        # The trailing unused fields may contain junk and must be ignored.
        mappings = []
        for _ in range(self.stream.getbuffer().nbytes // 16):
            src = self.read_double()
            tgt = self.read_double()
            mappings.append((src, tgt))

        return mappings


class MasterLocationParser(BaseParser):
    def _parse(self) -> tuple[int, tuple[float, float, float, float]]:
        # The location on all 4 axes for this master
        master_index = self.read_uint32()
        a1 = self.read_double()
        a2 = self.read_double()
        a3 = self.read_double()
        a4 = self.read_double()
        return master_index, (a1, a2, a3, a4)


class PrimaryInstancesParser(BaseParser):
    def _parse(self) -> list[PrimaryInstanceDict]:
        instances = []
        num_instances = self.read_value()
        for _ in range(num_instances):
            name = self.read_str_with_len()
            values = tuple(self.read_value() / 10000 for _ in range(4))
            instances.append(PrimaryInstanceDict(name=name, values=values))

        return instances



from enum import IntEnum


class F(IntEnum):
    BlockFileDataStart = 1
    BlockFontStart = 4
    FLVersion = 10
    BlockNamesStart = 262
    EncodingDefault = 1501
    Encoding = 1500
    MMEncType = 1502
    BlockNamesEnd = BlockNamesStart + 256
    BlockFontInfoStart = 257
    font_name = 1026
    MasterCount = 1503
    weight_vector = 1517
    unique_id = 1044
    version = 1046
    notice = 1038
    full_name = 1025
    family_name = 1027
    pref_family_name = 1024
    menu_name = 1056
    apple_name = 1092
    weight = 1028
    width = 1065
    License = 1069
    LicenseURL = 1070
    copyright = 1037
    trademark = 1061
    designer = 1062
    designer_url = 1063
    vendor_url = 1064
    source = 1039
    is_fixed_pitch = 1034
    weight_code = 1048
    italic_angle = 1029
    slant_angle = 1047
    underline_position = 1030
    underline_thickness = 1031
    ms_charset = 1054
    panose = 1118
    tt_version = 1128
    tt_u_id = 1129
    style_name = 1127
    pref_style_name = 1137
    mac_compatible = 1139
    SampleText = 1140
    vendor = 1121
    xuid = 1133
    xuid_num = 1134
    year = 1132
    version_major = 1130
    version_minor = 1131
    upm = 1135
    fond_id = 1090
    PostScriptHintingOptions = 1093
    Collection = 1068
    blue_values_num = 1530
    other_blues_num = 1531
    family_blues_num = 1532
    family_other_blues_num = 1533
    stem_snap_h_num = 1534
    stem_snap_v_num = 1535
    font_style = 1267
    pcl_id = 1057
    vp_id = 1058
    ms_id = 1060
    pcl_chars_set = 1059
    ttinfo = 1264
    unicoderanges = 2021
    fontnames = 1138
    CustomCMAPs = 1141
    PCLTTable = 1136
    ExportPCLTTable = 2022
    note = 2025
    FontFlags = 2030
    customdata = 2016
    MetricsClassFlags = 2024
    KerningClassFlags = 2026
    TrueTypeTable = 2014  # Font.truetypetables
    features = 1276
    GlyphClass = 1277  # Font.classes
    BlockFontInfoEnd = BlockFontInfoStart + 256
    BlockMMFontInfoStart = 271
    AxisCount = 1513
    AxisName = 1514
    AnisotropicInterpolationMappings = 1523
    AxisMappingsCount = 1515
    AxisMappings = 1516
    PrimaryInstanceLocations = 1247
    PrimaryInstances = 1254
    BlockMMFontInfoEnd = BlockMMFontInfoStart + 256
    # FIXME: Block start?
    GlobalGuides = 1294
    GlobalGuideProperties = 1296
    GlobalMask = 1295
    default_character = 1066
    # Glyph comes here
    FontOptions = 1743
    ExportOptions = 1744
    MappingMode = 1742
    # FIXME: Block end?
    BlockMMKerningStart = 272
    MMKernPair = 1410
    BlockMMKerningEnd = BlockMMKerningStart + 256
    BlockFontEnd = 5
    BlockFileDataEnd = 2

    # Place in order is unknown for those, because they didn't appear in our
    # test data:
    BlockIndex = 3
    BlockClipboardStart = 7
    BlockClipboardEnd = 8
    ClipboardFlags = 9
    BlockBlueInfoStart = 258
    BlockBlueInfoEnd = BlockBlueInfoStart + 256
    BlockOutlinesStart = 259
    BlockOutlinesEnd = BlockOutlinesStart + 256
    BlockGlobalHRStart = 260
    BlockGlobalHREnd = BlockGlobalHRStart + 256
    BlockSubrsStart = 261
    BlockSubrsEnd = BlockSubrsStart + 256
    BlockKerningStart = 263
    BlockKerningEnd = BlockKerningStart + 256
    BlockTrackingStart = 264
    BlockTrackingEnd = BlockTrackingStart + 256
    BlockEditorOptionsStart = 265
    BlockEditorOptionsEnd = BlockEditorOptionsStart + 256
    BlockDefEncodingStart = 266
    BlockDefEncodingEnd = BlockDefEncodingStart + 256
    BlockImportStart = 267
    BlockImportEnd = BlockImportStart + 256
    BlockMaskNames = 268
    BlockOptionsStart = 270
    BlockOptionsEnd = BlockOptionsStart + 256
    Bold = 1023
    FontMatrix = 1032
    FontBBox = 1033
    PaintType = 1035
    StrokeWidth = 1036
    PSAscent = 1040
    PSDescent = 1041
    PSCapHeight = 1042
    PSXHeight = 1043
    VMUsage = 1045
    XChar = 1050
    OChar = 1051
    SXChar = 1052
    SOChar = 1053
    MSFamily = 1055
    PSLeading = 1067
    BlueValues = 1100
    OtherBlues = 1101
    FamilyBlues = 1102
    FamilyOtherBlues = 1103
    BlueScale = 1104
    BlueShift = 1105
    BlueFuzz = 1106
    StdHW = 1107
    StdVW = 1108
    StemSnapH = 1109
    StemSnapV = 1110
    ForceBold = 1111
    LanguageGroup = 1112
    RndStemUp = 1113
    ExpansionFactor = 1114
    ErodeProc = 1115
    OtherSubrs = 1116
    LenIV = 1117
    IBMClass = 1119
    IBMSubclass = 1120
    Type1Map = 1126
    # Global hints and guides:
    GHRulers = 1280
    GVRulers = 1281
    GHHints = 1282
    GVHints = 1283
    HintReplace = 1290
    Hint3 = 1291
    ReplaceInfo = 1292
    ReplaceSet = 1293
    PGGuides = 1294
    PGlobalMask = 1295
    PGGuideNames = 1296
    # Subroutines:
    SubrNumber = 1300
    SubrName = 1301
    EX = 1302
    EY = 1303
    SubrBBox = 1304
    StartSubr = 1305
    EndSubr = 1306
    # Kerning:
    KernPair = 1400  # See also 1410 which we know
    # Tracking:
    TrackRecord = 1450
    # Encoding
    EncStandard = 1600
    AutoEncoding = 1603
    CustomDict = 2033
    AverageWidth = 1518
    # DefaultWidth = 1523  # Conflicts with AxisMap
    ForceBoldThreshold = 1519
    ReplaceTable = 1520
    Sample = 1521
    List = 1522
    OptionsTable = 1710
    OptionsMonitor = 1711
    OptionsTools = 1712
    OptionsLayers = 1713
    OptionsVectors = 1714
    OptionsList = 1715
    OptionsZoomFactor = 1720
    OptionsGrid = 1721
    OptionsGuides = 1722
    OptionsHints = 1723
    OptionsMask = 1724
    OptionsContour = 1725
    OptionsGridSize = 1726
    OptionsShowCross = 1727
    OptionsFitAscender = 1728
    OptionsFitDescender = 1729
    OptionsMaskColor = 1730
    OptionsInterfaceScale = 1731
    OptionsImportIndex = 1732
    OptionsExportIndex = 1733
    OptionsAssignMaskTo = 1734
    OptionsActiveLayer = 1740
    OptionsVisibleLayer = 1741


class G(IntEnum):
    Glyph = 2001
    Links = 2008  # Glyph.hlinks and Glyph.vlinks
    image = 2007
    Bitmaps = 2013
    VSB = 2023
    Sketch = 2019
    HintingOptions = 2010  # Better: Glyph Flags
    mask = 2009
    MaskMetrics = 2011  # was: mask.metrics
    MaskMetricsMM = 2028  # was: mask.metrics_mm
    Origin = 2027
    unicodes = 1250
    CustomDict = 2034
    UnicodesNonBMP = 1253
    mark = 2012
    customdata = 2015
    note = 2017
    GDEFData = 2018
    AnchorsProperties = 2020
    AnchorsMM = 2029
    GuideProperties = 2031
    # Place in order is unknown for those:
    CharIndes = 1200
    Contour = 1201
    Mask = 1202
    HRulers = 1203
    VRulers = 1204
    HHints = 1205
    VHints = 1206
    RPoints = 1207
    HGH = 1208
    VGH = 1209
    Base = 1210
    StartX = 1211
    StartY = 1212
    WidthX = 1213
    WidthY = 1214
    BBox = 1215
    LeftM = 1216
    RightM = 1217
    DIB = 1220
    DIBStartX = 1221
    DIBStartY = 1222
    Bmp = 1223
    Composite = 1224
    StartChar = 1230
    EndChar = 1231
    CharName = 1232
    SmallBmp = 1233
    StartCharSeg = 1240
    EndCharSeg = 1241
    # FontLab 3.0
    PContour = 2002
    PHints = 2003
    PGuides = 2004
    PMetrics = 2005
    EmptyGlyph = 2006
    LayersNum = 1235
    MMStartChar = 1236
    MMStartMaskChar = 1239
    MMWidth = 1237
    MMContour = 1238
    MMHHints = 1242
    MMVHints = 1243
    MMHGuides = 1244
    MMVGuides = 1245
    MMDHints = 1246
    MMMetrics = 1248
    NodesNum = 1249
    MMSeacs = 1251
    StyleTable = 1252


class M(IntEnum):
    MasterName = 1504
    MasterLocation = 1505
    PostScriptInfo = 1536


class T(IntEnum):
    cvt = 1261
    prep = 1262
    fpgm = 1263
    gasp = 1265
    vdmx = 1271
    hhea_line_gap = 1270
    hhea_ascender = 1278
    hhea_descender = 1279
    # hstem_data and vstem_data:
    TrueTypeStemPPEMs2And3 = 1266
    TrueTypeStemPPEMs = 1268
    TrueTypeStems = 1269
    TrueTypeStemPPEMs1 = 1524
    TrueTypeZones = 1255
    stemsnaplimit = 1272
    zoneppm = 1274
    codeppm = 1275
    dropoutppm = 1604
    MeasurementLine = 2032  # FIXME: Why in TrueType?
    TrueTypeZoneDeltas = 1273



from typing import Any

from vfbLib.compilers.base import BaseCompiler
from vfbLib.parsers.ps import global_options, glyph_options
from vfbLib.typing import GlyphHintingOptionsDict


class PostScriptInfoCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        font_matrix = data["font_matrix"]
        assert len(font_matrix) == 6, (
            f"'font_matrix' must have a length of 6, but has {len(font_matrix)}."
        )
        self.write_doubles(data["font_matrix"])
        self.write_int32(data["force_bold"])
        for k, size in (
            ("blue_values", 14),
            ("other_blues", 10),
            ("family_blues", 14),
            ("family_other_blues", 10),
        ):
            values = data[k]
            assert len(values) == size, (
                f"'{k}' must have a length of {size}, but has {len(values)}: {values}"
            )
            for value in values:
                self.write_int32(value)
        self.write_double(data["blue_scale"])
        for k in (
            "blue_shift",
            "blue_fuzz",
            "std_hw",
            "std_vw",
        ):
            self.write_uint32(data[k])
        for k, size in (
            ("stem_snap_h", 12),
            ("stem_snap_v", 12),
        ):
            values = data[k]
            assert len(values) == size, (
                f"'{k}' must have a length of {size}, but has {len(values)}."
            )
            for value in values:
                self.write_uint32(value)
        for k in ("xMin", "yMin", "xMax", "yMax"):
            self.write_int16(data["bounding_box"][k])
        for k in (
            "adv_width_min",
            "adv_width_max",
            "adv_width_avg",
            "ascender",
            "descender",
            "x_height",
            "cap_height",
        ):
            self.write_int32(data[k])


class PostScriptGlobalHintingOptionsCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        value = 0
        for k, bit in global_options:
            if data.get(k, 0):
                value += 2**bit
        for bit in data.get("other", []):
            value += 2**bit
        self.write_uint16(value)


class PostScriptGlyphHintingOptionsCompiler(BaseCompiler):
    def _compile(self, data: GlyphHintingOptionsDict) -> None:
        value = 0
        for k, bit in glyph_options:
            if data.get(k, 0):
                value += 2**bit
        for bit in data.get("other", []):
            value += 2**bit
        self.write_uint32(value)



import logging
from io import BytesIO
from typing import TYPE_CHECKING, Any

from vfbLib.compilers.header import VfbHeaderCompiler
from vfbLib.helpers import hexStr
from vfbLib.parsers.header import VfbHeaderParser

if TYPE_CHECKING:
    from io import BufferedIOBase

    from vfbLib.typing import VfbHeaderDict


logger = logging.getLogger(__name__)


class VfbHeader:
    def __init__(self) -> None:
        # The original or decompiled data
        self._data: "bytes | VfbHeaderDict | None" = None
        # The parser which can convert data to decompiled
        self.parser = VfbHeaderParser
        # The compiler which can convert the decompiled representation to bytes
        self.compiler = VfbHeaderCompiler

    def as_dict(self) -> dict[str, Any]:
        if isinstance(self.data, bytes):
            return {"decompiled": hexStr(self.data)}
        return {"decompiled": self.data}

    @property
    def data(self) -> "bytes | VfbHeaderDict | None":
        return self._data

    @data.setter
    def data(self, value: "bytes | VfbHeaderDict | None") -> None:
        self._data = value

    def compile(self) -> None:
        """
        Compile the header. The result is stored in VfbHeader.data.
        """
        if isinstance(self.data, bytes):
            return

        if self.data is None:
            raise TypeError("Can't compile empty header.")

        self.data = self.compiler().compile(self.data)

    def decompile(self) -> None:
        if not isinstance(self.data, bytes):
            # Already decompiled
            return

        byte_data = self.data
        self.data = self.parser(BytesIO(byte_data)).parse()

    def read(self, stream: "BufferedIOBase") -> None:
        # Read and decompile
        self.data = self.parser(stream).parse()



import logging
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any

from fontTools.designspaceLib import (
    AxisDescriptor,
    AxisLabelDescriptor,
    DesignSpaceDocument,
)
from fontTools.misc.textTools import deHexStr
from fontTools.ttLib.tables.ttProgram import Program, _indentRE, _unindentRE
from fontTools.ufoLib import UFOFileStructure
from ufoLib2.objects.features import Features
from ufoLib2.objects.font import Font
from ufoLib2.objects.glyph import Glyph
from ufonormalizer import normalizeUFO

from vfbLib.constants import ignore_minimal_keys
from vfbLib.enum import F, G, M, T
from vfbLib.typing import AnchorDict, GuidePropertiesDict, KerningClassFlagDict
from vfbLib.ufo.designspace import get_ds_location
from vfbLib.ufo.features import rename_kern_classes_in_feature_code
from vfbLib.ufo.glyph import VfbToUfoGlyph
from vfbLib.ufo.groups import transform_groups
from vfbLib.ufo.guides import apply_guide_properties, get_master_guides
from vfbLib.ufo.info import VfbToUfoInfo, default_tt_lib
from vfbLib.ufo.kerning import UfoKerning
from vfbLib.ufo.paths import UfoMasterGlyph
from vfbLib.ufo.tth import TTGlyphHints, transform_stem_rounds
from vfbLib.ufo.typing import (
    TUfoRawStemsDict,
    TUfoStemDict,
    TUfoStemPPMDict,
    TUfoStemPPMsDict,
    TUfoStemsDict,
    TUfoTTZoneDict,
    TUfoTTZonesDict,
)
from vfbLib.ufo.vfb2ufo import TT_GLYPH_LIB_KEY, TT_LIB_KEY, TT_UFO_LIB_KEY

if TYPE_CHECKING:
    from fontTools.designspaceLib import DiscreteAxisDescriptor

    from vfbLib.ufo.typing import UfoGroups, UfoMMKerning
    from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


class VfbToUfoBuilder:
    def __init__(
        self,
        vfb: "Vfb",
        minimal=False,
        base64=False,
        pshints=True,
        add_kerning_groups=False,
        move_groups=True,
        normalize=True,
    ) -> None:
        """Serialize the JSON structure to UFO(s)

        Args:
            vfb (Vfb): The `vfbLib.vfb.vfb.Vfb` to convert to UFO
            minimal (bool, optional): Convert only data necessary to generate font.
                Defaults to False.
            base64 (bool, optional): Store lib data in base64 encoding. Defaults to
                False.
            pshints (bool, optional): Include PostScript hinting. Defaults to True.
            add_kerning_groups (bool, optional): Add kerning groups to `features.fea`.
                Defaults to False.
            move_groups (bool, optional): Move non-kerning groups from `groups.plist` to
                `features.fea`. Defaults to True.
        """
        self.axes: "list[AxisDescriptor | DiscreteAxisDescriptor]" = []
        self.axis_count: int = 0
        self.vfb = vfb
        self.minimal = minimal
        self.encode_data_base64 = base64
        self.include_ps_hints = pshints
        self.add_kerning_groups_to_fea = add_kerning_groups
        self.move_groups_to_fea = move_groups
        self.normalize_ufos = normalize

        self.features_classes = ""
        self.features_code = ""
        self.groups: "UfoGroups" = {}
        self.guide_properties = GuidePropertiesDict(h=[], v=[])
        self.info = VfbToUfoInfo()
        self.kerning_class_flags: KerningClassFlagDict = KerningClassFlagDict()
        self.num_blue_values = 0
        self.num_other_blues = 0
        self.num_family_blues = 0
        self.num_family_other_blues = 0
        self.num_stem_snap_h = 0
        self.num_stem_snap_v = 0
        self.mm_guides = None
        self.mm_kerning: "UfoMMKerning" = {}
        self.kerning: dict[tuple[str, str], int] = {}
        self.lib: dict[str, Any] = {}
        self.masters: list[str] = []
        self.master_locations: dict[int, list[float]] = {}
        self.masters_ps_info: list[dict] = []
        self.primary_instances: list[dict[str, str | list[float]]] = []
        self.current_glyph: VfbToUfoGlyph | None = None
        self.glyph_masters: dict[str, VfbToUfoGlyph] = {}
        self.glyphOrder: list[str] = []
        # TT
        self.stem_ppms: TUfoStemPPMsDict = {"ttStemsH": [], "ttStemsV": []}
        self.stems: TUfoStemsDict = {"ttStemsH": [], "ttStemsV": []}
        self.tt_zones: TUfoTTZonesDict = {}
        self.tt_zone_names: list[str] = []
        self.zone_names: dict[str, list[str]] = {}
        self.build()

    def add_axis_mappings(self, data: list[tuple[float, float]]) -> None:
        if not self.axis_mappings_count:
            raise ValueError(
                "If axis mappings are present, axis mappings count must be set before "
                "parsing them."
            )
        for i in range(self.axis_count):
            # Get the number of mappings for the current axis
            n = self.axis_mappings_count[i]
            if n > 0:
                # Use only n mappings out of 10 for the current axis
                mappings = data[:n]
                axis = self.axes[i]
                axis.map = [(u, round(d * 1000)) for u, d in mappings]
                if isinstance(axis, AxisDescriptor):
                    # Derive min/max from the mappings
                    # FIXME: Is this info available elsewhere in the VFB?
                    user_coords = [c[0] for c in mappings]
                    axis.minimum = min(user_coords, default=0)
                    axis.default = axis.minimum
                    axis.maximum = max(user_coords, default=1000)
            # Jump to the next 10 axis mappings
            data = data[10:]

    def add_ot_class(self, data: str) -> None:
        """
        Add an OpenType class/group to the builder. VFB stores groups as strings, so
        minimal processing is done to extract the group name and the member glyph names.

        Args:
            data (str): The OpenType class from the VFB.
        """
        if ":" not in data:
            logger.warning(f"Malformed OT class definition, skipping: {data}")
            return

        name, contents = data.split(":")

        glyphs: list[str] = []
        for glyph in contents.split(" "):
            glyph = glyph.strip()
            if glyph:
                glyphs.append(glyph)

        if name in self.groups:
            logger.warning(f"Duplicate OT class name, skipping: {name}")
            return

        self.groups[name] = glyphs

    def assure_tt_lib(self) -> None:
        if TT_LIB_KEY not in self.lib:
            self.lib[TT_LIB_KEY] = {}

    def assure_tt_glyphlib(self) -> None:
        if TT_GLYPH_LIB_KEY not in self.lib:
            self.lib[TT_GLYPH_LIB_KEY] = {}

    def build_mm_glyph(self, data: dict[str, Any]) -> None:
        g = self.current_glyph = VfbToUfoGlyph(self)
        g.lib = {}
        g.name = data["name"]
        g.unicodes = []

        # MM Stuff, need to extract later
        if not self.minimal and "guides" in data:
            g.mm_guides = data["guides"]

        if self.include_ps_hints and "hints" in data:
            g.mm_hints = data["hints"]
            if "hintmasks" in data["hints"]:
                g.hintmasks = data["hints"]["hintmasks"]
                del data["hints"]["hintmasks"]

        if "kerning" in data:
            kerning: dict[int, list[int]] = data["kerning"]
            for Rid, values in kerning.items():
                assert g.name is not None
                self.mm_kerning[(g.name, Rid)] = values

        g.mm_metrics = data["metrics"]  # width and height
        g.mm_nodes = data["nodes"]

        if "components" in data:
            g.mm_components = data["components"]

        # TrueType hinting, stored here for later processing after all glyphs are there.
        if "tth" in data:
            g.tt_glyph_hints = TTGlyphHints(g, data["tth"], self.zone_names, self.stems)

    def set_feature_code(self, data: list[str]) -> None:
        # Make the kern feature compilable
        self.features_code = "\n".join(rename_kern_classes_in_feature_code(data))

    def set_glyph_background(self, data: dict[str, Any]) -> None:
        assert self.current_glyph is not None
        if "preview" in data["bitmap"]:
            del data["bitmap"]["preview"]
        self.current_glyph.lib["com.fontlab.v5.background"] = data

    def set_tt_cvt(self, data: str) -> None:
        from struct import iter_unpack

        values = [v[0] for v in iter_unpack("<h", deHexStr(data))]
        self.info.set_tt_instructions(
            "controlValue", {str(i): v for i, v in enumerate(values)}
        )

    def set_tt_program(self, key: str, data: bytes) -> None:
        p = Program()
        p.fromBytecode(deHexStr(data))
        # Apply indentation to imported binary program. Who said OCD?
        asm = "\n"
        white = "  "  # indent with two spaces
        indent = 0
        for line in p.getAssembly():
            if _unindentRE.match(line):
                indent -= 1
            asm += f"{white * indent}{line}\n"
            if _indentRE.match(line):
                indent += 1

        self.info.set_tt_instructions(key, asm)

    def set_tt_stem_ppms(self, data: TUfoStemPPMsDict) -> None:
        """Set the TT stem PPMs for stem widths of 2 to 5 pixels.

        Args:
            data (TUfoStemPPMsDict): The raw data.
        """
        for d in ("ttStemsH", "ttStemsV"):
            direction_stems = data[d]
            for ds in direction_stems:
                index = ds["stem"]
                # We can only show an index, self.stems is not filled yet
                name = f"{d[-1]}#{index}"
                rounds = transform_stem_rounds(ds["round"], name)
                stem: TUfoStemPPMDict = {"stem": index, "round": rounds}
                self.stem_ppms[d].append(stem)

    def set_tt_stem_ppms_1(self, data: TUfoStemPPMsDict) -> None:
        """Set the TT stem PPMs for stem width of 1 pixel.

        Args:
            data (TUfoStemPPMDict): The raw data.
        """
        for d in ("ttStemsH", "ttStemsV"):
            direction_stems = data[d]
            for ds in direction_stems:
                round_dict = self.stem_ppms[d][ds["stem"]]["round"]
                ppm = 1
                ppm1 = str(ds["round"][ppm])
                if ppm1 in round_dict:
                    logger.warning(
                        f"Duplicate rounding ppm {ppm1} in TT stem {d[-1]}#{ds['stem']}"
                        f", keeping value {round_dict[ppm1]}px over {ppm}px, "
                        f"which may not be what you want. {round_dict}"
                    )
                else:
                    round_dict[ppm1] = 1

    def set_tt_stems(self, data: TUfoRawStemsDict) -> None:
        for d in ("ttStemsH", "ttStemsV"):
            direction_stems = data[d]
            for i, ds in enumerate(direction_stems):
                # The rounding ppm value for 6 pixels is stored right here. We must
                #  re-add it to the dict later
                r = ds["round"]
                rk = list(r.keys())[0]
                rv = str(r[rk])

                # Take the other rounding ppm values (2-5) from self.stem_ppms
                stem_ppms = self.stem_ppms[d][i]
                assert i == stem_ppms["stem"]

                # Build the final stem dict
                stem: TUfoStemDict = {
                    "horizontal": d == "ttStemsV",
                    "width": ds["value"],
                    "name": ds["name"],
                    "round": stem_ppms["round"],
                }

                # Re-add the rounding ppm value for 6 pixels
                if rv in stem["round"]:
                    logger.warning(
                        f"Duplicate rounding ppm {rv} in TT stem {d[-1]}#{i}, "
                        f"choosing bigger value {rk}px over {stem['round'][rv]}px."
                    )
                stem["round"][rv] = int(rk)

                self.stems[d].append(stem)

    def set_tt_zones(self, data: dict[str, list]) -> None:
        self.zone_names = {
            "ttZonesT": [],
            "ttZonesB": [],
        }
        for d in ("ttZonesB", "ttZonesT"):
            direction_zones = data[d]
            for dz in direction_zones:
                zone = TUfoTTZoneDict(
                    position=dz["position"],
                    top=d == "ttZonesT",
                    width=dz["value"],
                )
                name = dz["name"]
                if name in self.tt_zones:
                    logger.warning(f"Duplicate zone name: {name}, overwriting.")
                self.tt_zones[name] = zone
                self.tt_zone_names.append(name)  # for deltas
                self.zone_names[d].append(name)  # for AlignTop/AlignBottom

    def set_tt_pixel_snap(self, data) -> None:
        self.assure_tt_lib()
        if data != 68:
            # Omit default value
            self.lib[TT_LIB_KEY]["stemsnaplimit"] = data

    def set_tt_zone_stop(self, data) -> None:
        self.assure_tt_lib()
        if data != 48:
            # Omit default value
            self.lib[TT_LIB_KEY]["zoneppm"] = data

    def set_tt_code_stop(self, data) -> None:
        self.assure_tt_lib()
        self.lib[TT_LIB_KEY]["codeppm"] = data

    def set_tt_dropout_stop(self, data) -> None:
        self.assure_tt_lib()
        if data != 255:
            # Omit default value
            self.lib[TT_LIB_KEY]["dropoutppm"] = data

    def set_tt_zone_deltas(self, data) -> None:
        for zone_index, deltas in data.items():
            zone_name = self.tt_zone_names[int(zone_index)]
            self.tt_zones[zone_name]["delta"] = {str(k): v for k, v in deltas.items()}

    def build_tt_stems_lib(self) -> None:
        lib = {}
        for d in ("ttStemsH", "ttStemsV"):
            i = 0
            direction_stems = self.stems[d]
            for stem in direction_stems:
                name = stem["name"]
                del stem["name"]
                if name == "":
                    dname = "v" if d == "ttStemsH" else "h"
                    name = "%s%02i" % (dname, i)
                    i += 1
                if name in lib:
                    oldname = name
                    name = "%s#%02i" % (oldname, i)
                    logger.warning(
                        f"Duplicate TrueType stem name '{oldname}', renamed to '{name}'"
                        ". Preferably make stem names unique in VFB."
                    )
                lib[name] = stem
        if lib:
            self.lib[TT_LIB_KEY]["stems"] = lib

    def build_tt_zones_lib(self) -> None:
        self.assure_tt_lib()
        if self.tt_zones:
            self.lib[TT_LIB_KEY]["zones"] = self.tt_zones

    def build_tt_glyph_lib(self) -> None:
        for glyph in sorted(self.glyph_masters.values()):
            if glyph.tt_glyph_hints is None:
                continue

            glyph.tth_commands = glyph.tt_glyph_hints.get_tt_glyph_hints()

    def build(self) -> None:  # noqa: C901
        # Non-MM data
        skip_keys = set(
            (
                F.Encoding,
                F.EncodingDefault,
                F.version,
                F.xuid,
                G.AnchorsProperties,
                G.HintingOptions,
                G.Origin,
            )
        )
        if self.vfb.minimal:
            skip_keys = skip_keys.union(ignore_minimal_keys)

        for e in self.vfb.entries:
            if e.key is None:
                raise TypeError

            key = e.id
            if key in skip_keys:
                continue

            data = e.data
            if data is None:
                continue

            # Font Info
            attr = self.info.mapping.get(key, None)
            if attr is not None:
                self.info.set_attr(attr, data)
                continue

            match key:
                case G.Glyph:
                    if self.current_glyph is not None:
                        name = self.current_glyph.name
                        if name in self.glyph_masters:
                            for i in range(255):
                                n = f"{name}#{i}"
                                if n not in self.glyph_masters:
                                    self.glyph_masters[n] = self.current_glyph
                                    self.glyphOrder.append(n)
                                    logger.warning(
                                        f"Duplicate glyph, renamed from {name} to {n}."
                                    )
                                    break
                        else:
                            if name is None:
                                logger.error("Glyph without name.")
                            else:
                                self.glyph_masters[name] = self.current_glyph
                                self.glyphOrder.append(name)
                    self.build_mm_glyph(data)
                case G.VSB:
                    assert self.current_glyph is not None
                    # TODO
                case G.unicodes:
                    assert self.current_glyph is not None
                    self.current_glyph.unicodes.extend(data)
                case G.UnicodesNonBMP:
                    assert self.current_glyph is not None
                    self.current_glyph.unicodes.extend(data)
                case G.Links:
                    assert self.current_glyph is not None
                    self.current_glyph.links = data
                case G.mark:
                    assert self.current_glyph is not None
                    self.current_glyph.set_mark(data)
                case G.customdata:
                    assert self.current_glyph is not None
                    self.current_glyph.lib["com.fontlab.v5.userData"] = data
                case G.note:
                    assert self.current_glyph is not None
                    self.current_glyph.note = data
                case G.GDEFData:
                    if "anchors" in data:
                        assert self.current_glyph is not None
                        self.current_glyph.anchors = []
                        for anchor in data["anchors"]:
                            a = AnchorDict(x=anchor["x"], y=anchor["y"])
                            if "name" in anchor:
                                a["name"] = anchor["name"]
                            self.current_glyph.anchors.append(a)
                case G.AnchorsMM:
                    assert self.current_glyph is not None
                    self.current_glyph.mm_anchors = data
                case G.GuideProperties:
                    assert self.current_glyph is not None
                    self.current_glyph.guide_properties = data
                case G.image:
                    self.set_glyph_background(data)
                case G.mask:
                    assert self.current_glyph is not None
                    self.current_glyph.set_mask(data)
                case G.MaskMetrics:
                    assert self.current_glyph is not None
                    w, h = data
                    self.current_glyph.mm_mask_metrics.append((w, h))
                case G.MaskMetricsMM:
                    assert self.current_glyph is not None
                    for w, h in data:
                        self.current_glyph.mm_mask_metrics.append((w, h))
                case F.pref_family_name:
                    self.info.familyName = data
                    self.info.openTypeNamePreferredFamilyName = data
                case F.is_fixed_pitch:
                    self.info.set_fixed_pitch(data)
                case F.weight_code:
                    self.info.set_weight_class(data)
                case F.style_name:
                    # FIXME: Does not contain Italic
                    self.info.postscriptWeightName = data
                case F.pref_style_name:
                    self.info.set_style_name(data)
                case F.fontnames:
                    self.info.set_name_records(data)
                case F.PrimaryInstances:
                    self.primary_instances = data
                case T.TrueTypeZones:
                    self.set_tt_zones(data)
                case F.ttinfo:
                    self.info.set_tt_info(data)
                case T.gasp:
                    self.info.set_tt_gasp(data)
                case F.font_style:
                    self.info.set_selection(data)
                case T.TrueTypeStemPPEMs:
                    self.set_tt_stem_ppms(data)
                case T.TrueTypeStems:
                    self.set_tt_stems(data)
                case T.stemsnaplimit:
                    self.set_tt_pixel_snap(data)
                case T.TrueTypeZoneDeltas:
                    self.set_tt_zone_deltas(data)
                case T.zoneppm:
                    self.set_tt_zone_stop(data)
                case T.codeppm:
                    self.set_tt_code_stop(data)
                case T.dropoutppm:
                    self.set_tt_dropout_stop(data)
                case F.features:
                    self.set_feature_code(data)
                case F.GlyphClass:
                    self.add_ot_class(data)
                case F.GlobalGuides:
                    self.mm_guides = data
                case F.GlobalGuideProperties:
                    self.guide_properties = data
                case F.MasterCount:
                    self.num_masters: int = data
                case M.MasterName:
                    self.masters.append(data)
                case M.MasterLocation:
                    index, locations = data
                    self.master_locations[index] = locations
                case F.AxisCount:
                    self.axis_count = data
                case F.AxisName:
                    tags = {
                        "Weight": "wght",
                        "Width": "wdth",
                        "Optical Size": "opsz",
                        "Serif": "SERF",
                    }
                    self.axes.append(
                        AxisDescriptor(tag=tags.get(data, data.upper()[:4]), name=data)
                    )
                case F.AxisMappingsCount:
                    self.axis_mappings_count: list[int] = data
                case F.AxisMappings:
                    self.add_axis_mappings(data)
                case F.AnisotropicInterpolationMappings:
                    # TODO: Can we properly output this to designspace?
                    for axis in data:
                        for src, tgt in axis:
                            if src != tgt:
                                maps = {
                                    self.axes[i].name: data[i]
                                    for i in range(self.axis_count)
                                }
                                logger.warning(
                                    "WARNING: Designspace output of anisotropic "
                                    "interpolation settings is not yet supported. You "
                                    "must set it up manually."
                                )
                                logger.warning(f"         Mappings: {maps}")
                                break
                case T.TrueTypeStemPPEMs1:
                    self.set_tt_stem_ppms_1(data)
                case F.blue_values_num:
                    self.num_blue_values = data
                case F.other_blues_num:
                    self.num_other_blues = data
                case F.family_blues_num:
                    self.num_family_blues = data
                case F.family_other_blues_num:
                    self.num_family_other_blues = data
                case F.stem_snap_h_num:
                    self.num_stem_snap_h = data
                case F.stem_snap_v_num:
                    self.num_stem_snap_v = data
                case M.PostScriptInfo:
                    self.masters_ps_info.append(data)
                case F.customdata:
                    self.lib["com.fontlab.v5.userData"] = data
                case F.unicoderanges:
                    self.info.openTypeOS2UnicodeRanges = data
                case F.MetricsClassFlags:
                    self.lib["com.fontlab.v5.metricsClassFlags"] = data
                case F.note:
                    self.info.note = data
                case F.KerningClassFlags:
                    self.kerning_class_flags = data
                case (
                    F.BlockFileDataStart
                    | F.BlockFontStart
                    | F.BlockNamesStart
                    | F.BlockNamesEnd
                    | F.BlockFontInfoStart
                    | F.BlockFontInfoEnd
                    | F.BlockMMFontInfoStart
                    | F.BlockMMFontInfoEnd
                    | F.BlockFontEnd
                    | F.BlockFileDataEnd
                ):
                    pass
                case T.cvt:
                    self.set_tt_cvt(data)
                case T.prep:
                    self.set_tt_program("controlValueProgram", data)
                case T.fpgm:
                    self.set_tt_program("fontProgram", data)
                case _:
                    k = key
                    for enum in (F, G, M, T):
                        try:
                            k = enum(key).name
                            break
                        except ValueError:
                            pass
                    logger.info(f"Unhandled entry: {k}")

        if self.current_glyph is not None:
            assert self.current_glyph.name is not None
            self.glyph_masters[self.current_glyph.name] = self.current_glyph
            self.glyphOrder.append(self.current_glyph.name)
        self.lib["public.glyphOrder"] = self.glyphOrder
        self.build_tt_glyph_lib()
        self.assure_tt_lib()
        self.build_tt_stems_lib()
        self.build_tt_zones_lib()
        if not self.lib[TT_LIB_KEY]:
            del self.lib[TT_LIB_KEY]
        if (
            TT_UFO_LIB_KEY in self.info.lib
            and self.info.lib[TT_UFO_LIB_KEY] == default_tt_lib
        ):
            del self.info.lib[TT_UFO_LIB_KEY]
        self.fix_masters_count()
        self.info.fix_underline_position()

    def fix_masters_count(self) -> None:
        # Sometimes no master name seems to be stored in the VFB. We need to add
        # synthetic names in this case.
        if self.axis_count == 0:
            masters_count = 1
        else:
            masters_count = self.axis_count**2
        if len(self.masters) < masters_count:
            additional = [f"m{n}" for n in range(len(self.masters), masters_count)]
            self.masters.extend(additional)

    def get_master_info(self, master_index: int = 0) -> VfbToUfoInfo:
        master_info = deepcopy(self.info)
        # Update the info with master-specific values
        properties = [
            ("ascender", "ascender"),
            ("descender", "descender"),
            ("x_height", "xHeight"),
            ("cap_height", "capHeight"),
        ]
        if self.include_ps_hints:
            properties.extend(
                [
                    # ("force_bold", "postscriptForceBold"),
                    ("blue_scale", "postscriptBlueScale"),
                    ("blue_shift", "postscriptBlueShift"),
                    ("blue_fuzz", "postscriptBlueFuzz"),
                ]
            )

        for key, attr in properties:
            value = self.masters_ps_info[master_index].get(key, None)
            if value is not None:
                setattr(master_info, attr, value)

        if self.include_ps_hints:
            # Set "force bold"
            value = self.masters_ps_info[master_index].get("force_bold", None)
            if value is not None:
                master_info.postscriptForceBold = bool(value)

            # Set attributes that must be clipped at a certain index
            for key, attr in (
                ("blue_values", "postscriptBlueValues"),
                ("other_blues", "postscriptOtherBlues"),
                ("family_blues", "postscriptFamilyBlues"),
                ("family_other_blues", "postscriptFamilyOtherBlues"),
                ("stem_snap_h", "postscriptStemSnapH"),
                ("stem_snap_v", "postscriptStemSnapV"),
            ):
                # Get the number of values from the corresponding attribute
                num_values = getattr(self, f"num_{key}")
                if num_values == 0:
                    continue

                value = self.masters_ps_info[master_index].get(key, None)
                if value is not None:
                    setattr(master_info, attr, value[:num_values])

        # Guides
        if not self.minimal and self.mm_guides is not None:
            guides = get_master_guides(self.mm_guides, master_index)
            apply_guide_properties(guides, self.guide_properties)

            if guides:
                master_info.guidelines = guides

        return master_info

    def get_ufo_master(self, index: int, silent=False) -> Font:
        if not silent:
            print(f"Processing font: {self.info.ui_name.strip()}, master {index}")

        # Build the master-specific data that can be passed when instantiating the UFO
        master_kerning = self.ufo_kerning.get_master_kerning(master_index=index)
        master_info = self.get_master_info(master_index=index)

        # Pass as much data right to the UFO
        ufo = Font(
            features=self.ufo_features.text,
            groups=deepcopy(self.ufo_groups),
            info=master_info,
            kerning=master_kerning,
            lib=self.lib,
        )

        mask_glyphs: list[Glyph] = []

        # Add the glyphs
        for name, mm_glyph in self.glyph_masters.items():
            logger.debug(f"    {name}, {type(name)}, {mm_glyph}")
            # TODO: Any way to use the ufoLib2 Glyph directly?
            master_glyph = UfoMasterGlyph(mm_glyph, self.glyphOrder, index)
            master_glyph.build(
                self.minimal, self.include_ps_hints, self.encode_data_base64
            )
            ufo_glyph = ufo.newGlyph(name)
            pen = ufo_glyph.getPointPen()
            master_glyph.drawPoints(pen)
            # FIXME: Anchors work, but typing wrong
            ufo_glyph.anchors = master_glyph.anchors
            ufo_glyph.guidelines = master_glyph.guidelines
            ufo_glyph.height = master_glyph.height
            ufo_glyph.lib = master_glyph.lib
            ufo_glyph.width = master_glyph.width
            ufo_glyph.unicodes = master_glyph.unicodes

            if self.minimal:
                continue

            # Handle mask layer
            mask_ufo_glyph = Glyph(name)
            pen = mask_ufo_glyph.getPointPen()
            master_glyph.drawPointsMask(pen)
            mask_width, mask_height = master_glyph.mask_metrics
            mask_ufo_glyph.width = mask_width
            mask_ufo_glyph.height = mask_height

            if mask_ufo_glyph.contours:
                # Only include mask glyphs with contours. There is bogus width
                # information in empty mask glyphs in the VFB.
                mask_glyphs.append(mask_ufo_glyph)

        if mask_glyphs:
            # Include "mask" as background layer
            mask = ufo.layers.newLayer("public.background")
            # Add the non-empty mask glyphs
            for mask_glyph in mask_glyphs:
                mask.addGlyph(mask_glyph)

        return ufo

    def get_ufo_masters(self, silent=False) -> list[Font]:
        # Prepare data shared by the master UFOs
        ufo_groups, group_order, key_glyphs = transform_groups(
            self.groups,
            self.kerning_class_flags,
            self.glyphOrder,
            self.minimal,
        )
        self.ufo_kerning = UfoKerning(
            self.glyphOrder, ufo_groups, self.mm_kerning, key_glyphs
        )
        self.ufo_groups = self.ufo_kerning.groups

        if group_order:
            self.lib["public.groupOrder"] = group_order
        if key_glyphs:
            self.lib["com.lucasfonts.vfblib.groupKeyGlyphs"] = key_glyphs
        if self.info.lib:
            self.lib.update(self.info.lib)

        self.ufo_features = Features()

        # Also add non-kerning classes to the feature code
        move_groups = set()
        for name, glyphs in ufo_groups.items():
            if name.startswith("public.kern") and not self.add_kerning_groups_to_fea:
                # Kerning group
                continue
            if name.startswith("."):
                # Metrics group
                continue
            if not name.startswith("public.kern") and not self.move_groups_to_fea:
                # OT group
                continue

            self.ufo_features.text += f"@{name} = [{' '.join(glyphs)}];\n"
            move_groups.add(name)

        if self.move_groups_to_fea:
            # Remove groups that were added to the feature code from the groups dict
            for name in move_groups:
                try:
                    del self.ufo_groups[name]
                except KeyError:
                    logger.warning(f"Trying to remove non-existing group: {name}")

        if self.features_code:
            self.ufo_features.text += f"\n\n{self.features_code}"

        ufo_masters = []
        logger.info(f"Extracting {len(self.masters)} master UFOs...")
        for i in range(len(self.masters)):
            ufo_masters.append(self.get_ufo_master(i, silent))
        return ufo_masters

    def get_ufos_designspace(
        self, out_path: Path, silent=False
    ) -> tuple[list[Font], DesignSpaceDocument | None]:
        """
        Build UFOs and a DesignSpaceDocument from the VFB contents in memory and return
        them. The DesignSpaceDocument is only returned for VFBs containing more than one
        master. In other cases, the second element of the returned tuple is None.
        """
        ufo_masters = self.get_ufo_masters(silent)
        if not ufo_masters:
            logger.error("Could not extract any UFO masters.")
            raise ValueError

        ds: DesignSpaceDocument | None = None
        if self.axes:
            ds = self.get_designspace(out_path)
        return ufo_masters, ds

    def write(
        self, out_path: Path, overwrite=False, silent=False, ufoz=False, json=False
    ) -> None:
        """
        Write a the VFB contents to master UFOs and a designspace file. The designspace
        file is only written if the VFB contains more than one master.
        """
        if json:
            out_path = out_path.with_suffix(".json")
        # Build UFOs and DesignSpace
        ufos, ds = self.get_ufos_designspace(out_path, silent)

        # Write the master UFOs
        strct = UFOFileStructure.ZIP if ufoz else None
        indent = None if ufoz else 2
        for index, ufo in enumerate(ufos):
            master_path = self.get_master_path(out_path, index)
            if json:
                ufo.json_dump(master_path, indent=indent, sort_keys=False)
            else:
                ufo.save(master_path, structure=strct, overwrite=overwrite)
                if self.normalize_ufos and not ufoz:
                    normalizeUFO(
                        ufoPath=master_path, onlyModified=False, writeModTimes=False
                    )

        # Write the Designspace
        if ds:
            ds_path = out_path.with_suffix(".designspace")
            if ds_path.exists():
                if not overwrite:
                    raise FileExistsError(str(ds_path))
            if not silent:
                print(f"Writing designspace: {ds_path}")
            ds.write(str(ds_path))

    def get_designspace(self, out_path: Path) -> DesignSpaceDocument:
        """
        Build and return a DesignSpaceDocument. The out_path argument will be used to
        construct the UFO paths assigned to the source descriptors.
        """
        ds = DesignSpaceDocument()
        ds.axes = self.axes

        # Add labels
        if self.axis_count == 1:
            # Only support the simple case here
            for p in self.primary_instances:
                self.axes[0].axisLabels.append(
                    AxisLabelDescriptor(name=p["name"], userValue=p["values"][0])
                )
        else:
            logger.warning(
                "WARNING: Designspace output of AxisLabelDescriptors is only supported "
                "for VFBs with 1 axis. Please add labels manually."
            )

        # Add sources
        for i in range(self.num_masters):
            ds.addSourceDescriptor(
                # designLocation=,
                familyName=self.info.ds_family_name,
                filename=f"{self.get_master_path(out_path, i)}",
                location=get_ds_location(self.axes, self.master_locations[i + 1]),
                name=self.masters[i].strip(),
                styleName=f"Master {i}",
            )

        # Add instances
        for i in range(len(self.primary_instances)):
            p = self.primary_instances[i]
            loc = p["values"]
            if not (isinstance(loc, list) or isinstance(loc, tuple)):
                raise TypeError(f"Expected list or tuple, got {type(loc)}")

            style_name = p["name"]
            # FIXME: Support anisotropic interpolation
            ds.addInstanceDescriptor(
                # designLocation=get_ds_design_location(self.axes, loc),
                familyName=self.info.ds_family_name,
                filename=f"instance_ufo/{self.info.ds_family_name}-{style_name}.ufo",
                # lib=,
                # locationLabel="",
                name=f"instance_{i}",
                # postScriptFontName="",
                # FIXME:
                styleMapFamilyName=f"{self.info.ds_family_name} {style_name}".strip(),
                styleMapStyleName=self.info.styleMapStyleName,
                styleName=style_name,
                userLocation=get_ds_location(self.axes, loc, 1),
            )

        return ds

    @classmethod
    def get_master_path(cls, out_path: Path, master_index: int) -> Path:
        if master_index > 0:
            try:
                return out_path.with_stem(f"{out_path.stem}-{master_index}")
            except AttributeError:
                # Python 3.8 doesn't have `Path.with_stem`
                return out_path.with_name(
                    f"{out_path.stem}-{master_index}{out_path.suffix}"
                )
        return out_path



import logging
from html import escape as html_escape
from typing import TYPE_CHECKING, Any

from vfbLib.ufo.glyph import VfbToUfoGlyph
from vfbLib.ufo.vfb2ufo import (
    TT_GLYPH_LIB_KEY,
    vfb2ufo_alignment_rev,
    vfb2ufo_command_codes,
)

if TYPE_CHECKING:
    from vfbLib.ufo.typing import TUfoStemsDict


logger = logging.getLogger(__name__)


def get_xml_tth(commands) -> list[str]:
    """
    Convert the glyph's list of TTH commands to a list of TTH command xml strings.
    """
    return [tt_cmd_dict_to_xml(cmd_dict) for cmd_dict in commands]


def set_tth_lib(glyph, commands) -> None:
    """
    Save the TTH commands to the glyph's lib. Optionally rename the hinted points.
    """
    tth = get_xml_tth(commands)
    if tth:
        glyph.lib[TT_GLYPH_LIB_KEY] = (
            "  <ttProgram>\n" + "\n".join(tth) + "\n  </ttProgram>\n"
        )


def tt_cmd_dict_to_xml(tt_dict: dict[str, Any]) -> str:
    """
    Convert the dict tt command into a FontLab XML string.
    """
    code = tt_dict["code"]
    cmd = f'    <ttc code="{code}"'
    for attr in (
        "point",
        "point1",
        "point2",
        "round",
        "stem",
        "zone",
        "align",
        "delta",
        "ppm1",
        "ppm2",
    ):
        if attr in tt_dict:
            if attr == "round":
                val = str(tt_dict[attr]).lower()
            else:
                val = tt_dict[attr]
            cmd += f' {attr}="{html_escape(str(val))}"'
    cmd += "/>"
    return cmd


def transform_stem_rounds(data: dict[int, int], name: str) -> dict[str, int]:
    """Transform the format of the stem rounding dict to fit the UFO output format, i.e.
    exchange key and value.

    Args:
        data (dict[int, int]): The stem rounding data
        name (str): A name that will be shown if there is any error.

    Returns:
        dict[str, int]: The transformed stem rounding dict.
    """
    d: dict[str, int] = {}
    for k, v in data.items():
        key = v
        val = k
        if key in d:
            if val > v:
                logger.warning(
                    f"Duplicate rounding ppm {key} in TT stem '{name}', "
                    f"choosing bigger value {val}px over {d[key]}px. {data}"
                )
                d[key] = val
            else:
                logger.warning(
                    f"Duplicate rounding ppm {key} in TT stem '{name}', "
                    f"keeping value {d[key]}px, ignoring {val}px. {data}"
                )
        else:
            d[str(key)] = val
    return d


class TTGlyphHints:
    def __init__(
        self,
        mm_glyph: VfbToUfoGlyph,
        data: list[dict[str, Any]],
        zone_names: dict[str, list[str]],
        stems: "TUfoStemsDict",
    ) -> None:
        self.glyph: VfbToUfoGlyph = mm_glyph
        self.data = data
        self.zone_names = zone_names
        self.stems = stems

    def get_tt_glyph_hints(self) -> list[dict[str, str | bool]]:
        # Build TT hints which into glyph lib and return them.
        commands: list[dict[str, str | bool]] = []
        for cmd in self.data:
            code = cmd["cmd"]
            params = cmd["params"]
            d: dict[str, str | bool] = {"code": vfb2ufo_command_codes[code]}
            if code in ("AlignBottom", "AlignTop"):
                d["point"] = self.glyph.get_point_label(params["pt"], code)
                if code == "AlignBottom":
                    zd = "ttZonesB"
                else:
                    zd = "ttZonesT"
                # Zones are indexed separately for top and bottom sides
                d["zone"] = self.zone_names[zd][params["zone"]]
            elif code in ("AlignH", "AlignV"):
                d["point"] = self.glyph.get_point_label(params["pt"], code)
                if "align" in params:
                    align = params["align"]
                    if align > -1:
                        d["align"] = vfb2ufo_alignment_rev.get(align, "round")
            elif code in (
                "SingleLinkH",
                "SingleLinkV",
                "DoubleLinkH",
                "DoubleLinkV",
            ):
                d["point1"] = self.glyph.get_point_label(params["pt1"], code)
                d["point2"] = self.glyph.get_point_label(params["pt2"], code)
                if "stem" in params:
                    stem = params["stem"]
                    if stem <= -2:
                        d["round"] = True
                    elif stem == -1:
                        pass
                    else:
                        stem_dir = "ttStemsH" if code.endswith("H") else "ttStemsV"
                        # Stems are indexed separately for H and V directions
                        if stem >= len(self.stems[stem_dir]):
                            logger.warning(
                                f"Stem index in {stem_dir} out of range in "
                                f"{self.glyph.name}: {stem} (of "
                                f"{len(self.stems[stem_dir])} existing stems). "
                                "Choosing first stem."
                            )
                            logger.warning(f"{code}: {params}")
                            logger.warning(self.stems[stem_dir])
                            stem = 0
                        d["stem"] = self.stems[stem_dir][stem]["name"]
                if "align" in params:
                    align = params["align"]
                    if align > -1:
                        d["align"] = vfb2ufo_alignment_rev.get(align, "round")
            elif code in (
                "InterpolateH",
                "InterpolateV",
            ):
                d["point"] = self.glyph.get_point_label(params["pti"], code)
                d["point1"] = self.glyph.get_point_label(params["pt1"], code)
                d["point2"] = self.glyph.get_point_label(params["pt2"], code)
                if "align" in params:
                    align = params["align"]
                    if align > -1:
                        d["align"] = vfb2ufo_alignment_rev.get(align, "round")
            elif code in (
                "MDeltaH",
                "MDeltaV",
                "FDeltaH",
                "FDeltaV",
            ):
                d["point"] = self.glyph.get_point_label(params["pt"], code)
                d["delta"] = params["shift"]
                d["ppm1"] = params["ppm1"]
                d["ppm2"] = params["ppm2"]
            else:
                logger.error(f"Unknown TT command: {code}")
                raise ValueError

            commands.append(d)
        return commands



import logging
from typing import TYPE_CHECKING

from vfbLib.ufo.groups import build_glyph_to_group_maps

if TYPE_CHECKING:
    from vfbLib.ufo.typing import UfoGroups, UfoMasterKerning, UfoMMKerning


logger = logging.getLogger(__name__)


class UfoKerning:
    def __init__(
        self,
        glyph_order: list[str],
        groups: "UfoGroups",
        mm_kerning: "UfoMMKerning",
        key_glyphs: dict[str, str],
    ):
        self.glyph_order = glyph_order
        group_info = build_glyph_to_group_maps(groups)
        self.groups, self.glyph_group_1, self.glyph_group_2 = group_info
        self.mm_kerning = mm_kerning
        self.key_glyphs = self._reverse_key_glyph_dict(key_glyphs)
        self._make_name_based_kerning()

    def _is_exception(self, L: str, R: str):
        L_is_key = (1, L) in self.key_glyphs or L not in self.glyph_group_1

        R_is_key = (2, R) in self.key_glyphs or R not in self.glyph_group_2

        if L_is_key and R_is_key:
            return False

        return True

    def _make_name_based_kerning(self) -> None:
        """
        Convert the glyph indices to glyph names. Also solves group kerning references.
        """
        self.mm_kerning_names: dict[tuple[str, str], list[int]] = {}
        for pair, values in self.mm_kerning.items():
            L, Rid = pair
            # Make right GID into glyph name
            R = self.glyph_order[int(Rid)]

            # Is the left glyph a keyglyph? In that case, use the group name instead of
            # the glyph name.
            left_group = self.key_glyphs.get((1, L))
            left = left_group if left_group in self.groups else L

            # Is the right glyph a keyglyph?
            right_group = self.key_glyphs.get((2, R))
            right = right_group if right_group in self.groups else R

            self.mm_kerning_names[left, right] = values

    def _reverse_key_glyph_dict(
        self, key_glyphs: dict[str, str]
    ) -> dict[tuple[int, str] | None, str]:
        """
        Rebuild the key glyphs dict (group_name: key_ glyph_name) into a reverse dict by
        kerning group side and key glyph to group name.
        """
        rev = {}
        for name, key_glyph in key_glyphs.items():
            if name.startswith("public.kern1"):
                key = (1, key_glyph)
            elif name.startswith("public.kern2"):
                key = (2, key_glyph)
            else:
                key = None
                logger.warning(
                    f"Can't determine kerning group side from name: '{name}'"
                )
            if key in rev:
                logger.warning(
                    f"Glyph is key glyph for more than one kerning group: {key}"
                )
                continue

            rev[key] = name
        return rev

    def get_master_kerning(self, master_index: int) -> "UfoMasterKerning":
        """
        Extract the kerning values for master_index and return the kerning as
        dict[tuple[str, str], int].
        """
        master_kerning: "UfoMasterKerning" = {}
        for pair, values in self.mm_kerning_names.items():
            value = values[master_index]
            if value != 0 or self._is_exception(*pair):
                master_kerning[pair] = value
        return master_kerning



import logging
from typing import TYPE_CHECKING

from vfbLib import DIRECTIONS
from vfbLib.ufo.typing import UfoGuide

if TYPE_CHECKING:
    from vfbLib.typing import GuidePropertiesDict, MMGuidesDict


logger = logging.getLogger(__name__)


def get_master_guides(mm_guides: "MMGuidesDict", master_index: int) -> list[UfoGuide]:
    # Concatenate guidlines for both directions and extract coords for
    # master_index
    guides = []
    for direction in DIRECTIONS:
        for direction_mm_guide in mm_guides[direction]:
            master_guide = direction_mm_guide[master_index]
            guide = UfoGuide(angle=0, x=0, y=0, _direction=direction)
            if direction == "h":
                guide["y"] = master_guide["pos"]
            else:
                guide["x"] = master_guide["pos"]
            angle = master_guide["angle"]
            if direction == "v":
                angle = 90 - angle
            if angle < 0:
                angle += 360
            if angle > 360:
                angle -= 360
            guide["angle"] = round(angle, 2)
            guides.append(guide)
    return guides


def apply_guide_properties(
    guides: list[UfoGuide], properties: "GuidePropertiesDict"
) -> None:
    # Update the guides with names and colors from properties

    # Split the guides into h and v again
    direction_guides = {
        "h": [g for g in guides if g["_direction"] == "h"],
        "v": [g for g in guides if g["_direction"] == "v"],
    }

    for direction in DIRECTIONS:
        for prop in properties[direction]:
            guide_index = prop["index"] - 1  # index is 1-based
            if guide_index >= len(direction_guides[direction]):
                logger.info(
                    f"Skipping guide properties, guide is missing ({guide_index})."
                )
                logger.info(prop)
                logger.info(f"{direction}: {direction_guides[direction]}")
                continue

            guide = direction_guides[direction][guide_index]
            if "color" in prop:
                color = prop["color"]
                r = int(color[1:3], 16) / 0xFF
                try:
                    g = int(color[3:5], 16) / 0xFF
                except ValueError:
                    g = 0
                try:
                    b = int(color[5:7], 16) / 0xFF
                except ValueError:
                    b = 0
                guide["color"] = f"{r:0.4f},{g:0.4f},{b:0.4f},1"
            if "name" in prop:
                guide["name"] = prop["name"]
    for guide in guides:
        del guide["_direction"]



from collections.abc import Sequence
from enum import IntEnum
from typing import Literal

# Used by guides and links
DIRECTIONS: Sequence[Literal["h", "v"]] = ("h", "v")

GLYPH_CONSTANT = (1, 9, 7, 1)

export_options = {
    0: "use_custom_opentype_export_options",
    1: "use_default_opentype_export_options",  # FIXME: Correct?
    2: "use_custom_cmap_encoding",
}

gdef_class_names = (
    "unassigned",  # 0
    "base",  # 1
    "ligature",  # 2
    "mark",  # 3
    "component",  # 4
)

replace_types = {
    0x01: "h",  # hintmask for hstem
    0x02: "v",  # hintmask for vstem
    0xFF: "r",  # Replacement point
    # FIXME: This seems to be the node index of the replacement
    # point. But sometimes it is negative, why?
}

mapping_modes = {
    0: "names_or_index",
    1: "unicode_ranges",
    3: "codepages",
}

font_options = {
    1: "fit_ascender",
    2: "fit_descender",
    3: "auto_metrics_left",
    4: "auto_metrics_right",
    5: "auto_metrics_close",
    6: "auto_hinting_min_h_len",
    7: "auto_hinting_min_v_len",
    8: "auto_hinting_min_h_width",
    9: "auto_hinting_min_v_width",
    10: "auto_hinting_max_h_width",
    11: "auto_hinting_max_v_width",
    12: "auto_hinting_h_ratio",  # when reading, divide by 10000
    13: "auto_hinting_v_ratio",  # when reading, divide by 10000
    14: "duplicate_place_x",
    15: "paste_place_x",
    16: "opentype_name_records",
    #    0 - "append_opentype_records_to_default_names"
    #    1 - "do_not_export_opentype_name_records"
    #    2 - "export_only_opentype_name_records"
    17: "codepage_for_cmap_1_0",
    #    0 - MacRoman (default)
    #    1 - Current codepage in font window
    #    2 - Windows 1250 Central European
    # Didn't bother to check from here on:
    #    3 - Windows 1251 Cyrillic
    #    4 - Windows 1252 Western (ANSI)
    #    ...
    #  153 - NeXT OS NextStep Multinational
    18: "dont_ignore_unicode_indexes",
    19: "head_bbox_savings",
    20: "autohinting_options",  # Bit field, see below
    21: "export_hinted_truetype_font",
    22: "autohint_unhinted_glyphs",
    23: "keep_existing_truetype_instructions",
    24: "export_visual_truetype_hints",
    25: "apply_bbox_savings",
    26: "auto_win_asc_desc",
    27: "add_characters",
    28: "export_embedded_bitmaps",
    29: "copy_hdmx_data_from_base_to_composite_glyph",
    30: "dont_automatically_reorder_glyphs",
    31: "export_ot",
    32: "export_volt",
    33: "write_kern_feature",
    34: "t1_terminal",
    35: "t1_pfm",
    36: "t1_afm",
    37: "t1_autohint",
    38: "t1_unicode",
    39: "optimize_align",
    40: "optimize_reduce",
    41: "t1_encoding",
    42: "ot_write_gdef",
    43: "t1_use_os2",
    44: "subrize",
    45: "t1_sort",
    46: "export_kern_table",
    47: "t1_fs_type",
    48: "expand_kern_flags",  # Bit field, see below
    49: "expand_kern_codepage",
    50: "expand_kern_count",
    51: "decompose",
    # 100: "end",
    114: "duplicate_place_y",
    115: "paste_place_y",
}


class TTAutoHintOptions(IntEnum):
    single_link_attachment_precision = 0x7
    generate_triple_hints = 0x100
    generate_delta_instructions = 0x200
    direct_links_to_center_of_the_glyph_where_possible = 0x400
    interpolate_positions_of_cusp_points = 0x800
    interpolate_positions_of_double_links = 0x1000
    add_link_to_rsb = 0x2000


class ExpandKernOptions(IntEnum):
    limit_action = 0x1
    limit_codepage = 0x2
    limit_cmap_10 = 0x4
    limit_font_window = 0x8
    limit_count = 0x10
    limit_keep = 0x20
    apply_to_assistance = 0x1000


ttinfo_names = {
    # 0x32: "end",
    0x33: "max_zones",
    0x34: "max_twilight_points",
    0x35: "max_storage",
    0x36: "max_function_defs",
    0x37: "max_instruction_defs",
    0x38: "max_stack_elements",
    0x39: "head_flags",  # tt_font_info_settings with head flags combined
    0x3A: "head_units_per_em",  # units_per_em, duplicate
    0x3B: "head_mac_style",
    0x3C: "head_lowest_rec_ppem",  # lowest_rec_ppem
    0x56: "head_creation",  # timestamp
    0x57: "head_creation2",  # head_modification? returned as second value in list
    0x3D: "head_font_direction_hint",  # font_direction_hint
    0x3E: "os2_us_weight_class",  # weight_class, duplicate
    0x3F: "os2_us_width_class",  # width_class, duplicate
    0x40: "os2_fs_type",  # embedding
    0x41: "os2_y_subscript_x_size",  # subscript_x_size
    0x42: "os2_y_subscript_y_size",  # subscript_y_size
    0x43: "os2_y_subscript_x_offset",  # subscript_x_offset
    0x44: "os2_y_subscript_y_offset",  # subscript_y_offset
    0x45: "os2_y_superscript_x_size",  # superscript_x_size
    0x46: "os2_y_superscript_y_size",  # superscript_y_size
    0x47: "os2_y_superscript_x_offset",  # superscript_x_offset
    0x48: "os2_y_superscript_y_offset",  # superscript_y_offset
    0x49: "os2_y_strikeout_size",  # strikeout_size
    0x4A: "os2_y_strikeout_position",  # strikeout_position
    0x4B: "os2_s_family_class",  # ibm_classification + subclass
    0x4C: "OpenTypeOS2Panose",
    0x4D: "os2_s_typo_ascender",  # OpenTypeOS2TypoAscender
    0x4E: "os2_s_typo_descender",  # OpenTypeOS2TypoDescender
    0x4F: "os2_s_typo_line_gap",  # OpenTypeOS2TypoLineGap
    0x50: "os2_fs_selection",
    0x51: "os2_us_win_ascent",  # OpenTypeOS2WinAscent
    0x52: "os2_us_win_descent",  # OpenTypeOS2WinDescent
    0x5C: "Average Width",
    0x53: "Hdmx PPMs 1",
    0x58: "Hdmx PPMs 2",
    # os2_ul_code_page_range1, os2_ul_code_page_range2:
    0x54: "Codepages",
    # Position unknown:
    0x59: "hhea_line_gap",
    0x5A: "hhea_ascender",
    0x5B: "hhea_descender",
}

tt_settings = {
    0: "use_custom_tt_values",
    1: "create_vdmx",
    2: "add_null_cr_space",
}


class TTFlags(IntEnum):
    use_custom_tt_values = 0x00010000
    create_vdmx = 0x00020000
    add_null_cr_space = 0x00040000
    create_hdmx = 0x00080000
    create_vmtx = 0x00100000



from typing import Any

from vfbLib.compilers.base import BaseCompiler


class NameRecordsCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        self.write_value(len(data))  # number of records
        for nr in data:
            nameID, platID, encID, langID, name = nr
            self.write_value(nameID)
            self.write_value(platID)
            self.write_value(encID)
            self.write_value(langID)
            self.write_value(len(name))
            for char in name:
                if platID == 1 and encID == 0:
                    try:
                        char = char.encode("macroman")
                    except ValueError:
                        char = "?"
                o = ord(char)
                self.write_value(o)


class OpenTypeStringCompiler(BaseCompiler):
    """
    A compiler that compiles string data that represents OpenType feature code.
    """

    def _compile(self, data: Any) -> None:
        self.write_str("\n".join(data))


class StringCompiler(BaseCompiler):
    """
    A compiler that compiles string data.
    """

    def _compile(self, data: Any) -> None:
        self.write_str(data)


class VendorIdCompiler(BaseCompiler):
    """
    A compiler that compiles string data, padded to 4 bytes.
    """

    def _compile(self, data: Any) -> None:
        self.write_str(data, pad=4)



import logging
import xml.etree.ElementTree as elementTree
from typing import TYPE_CHECKING

from vfbLib import DIRECTIONS
from vfbLib.ufo.typing import HintSet, UfoHintingV2, UfoHintSet
from vfbLib.ufo.vfb2ufo import PS_GLYPH_LIB_KEY

if TYPE_CHECKING:
    from vfbLib.typing import HintDict, HintTuple
    from vfbLib.ufo.glyph import VfbToUfoGlyph
    from vfbLib.ufo.paths import UfoMasterGlyph


logger = logging.getLogger(__name__)


def normalize_hint(hint: tuple[str, int, int]):
    direction, pos, width = hint
    if width < 0:
        if width not in (-21, -20):  # Skip ghost hints
            pos = pos + width
            width = abs(width)
    return (direction, pos, width)


def normalize_hint_dict(hint: "HintDict", name: str = "dummy"):
    return normalize_hint((name, hint["pos"], hint["width"]))


def build_ps_glyph_hints(
    mmglyph: "VfbToUfoGlyph",
    glyph: "UfoMasterGlyph",
    master_hints: "dict[str, list[HintTuple]]",
) -> None:
    """
    Set the master-specific hints from data to the glyph lib

    Args:
        mmglyph (VfbToUfoGlyph): The multiple-master auxilliary glyph object
        glyph (UfoMasterGlyph): The target glyph
        master_hints (dict[str, list[HintTuple]]): The hints with their position and
            width adapted to the master
    """
    # Set the master-specific hints from data to the glyph lib
    # Use the format defined in UFO3, not what FL does.
    # https://unifiedfontobject.org/versions/ufo3/glyphs/glif/#publicpostscripthints

    # Quote from
    # https://github.com/adobe-type-tools/psautohint/blob/master/python/psautohint/ufoFont.py
    """
    A <dict> element in the hintSetList array identifies a specific point by its
    name, and describes a new set of stem hints which should be applied before the
    specific point.

    A <string> element in the flexList identifies a specific point by its name.
    The point is the first point of a curve. The presence of the element is a
    processing suggestion, that the curve and its successor curve should be
    converted to a flex operator.

    One challenge in applying the hintSetList and flexList elements is that in
    the GLIF format, there is no explicit start and end operator: the first path
    operator is both the end and the start of the path. I have chosen to convert
    this to T1 by taking the first path operator, and making it a move-to. I then
    also use it as the last path operator. An exception is a line-to; in T1, this
    is omitted, as it is implied by the need to close the path. Hence, if a hintset
    references the first operator, there is a potential ambiguity: should it be
    applied before the T1 move-to, or before the final T1 path operator? The logic
    here applies it before the move-to only.
    <glyph>
    ...
        <lib>
            <dict>
                <key><com.adobe.type.autohint></key>
                <dict>
                    <key>id</key>
                    <string> <fingerprint for glyph> </string>
                    <key>hintSetList</key>
                    <array>
                        <dict>
                        <key>pointTag</key>
                        <string> <point name> </string>
                        <key>stems</key>
                        <array>
                            <string>hstem <position value> <width value></string>*
                            <string>vstem <position value> <width value></string>*
                            <string>hstem3 <position value 0>...<position value 5>
                            </string>*
                            <string>vstem3 <position value 0>...<position value 5>
                            </string>*
                        </array>
                        </dict>*
                    </array>

                    <key>flexList</key>*
                    <array>
                        <string><point name></string>+
                    </array>
                </dict>
            </dict>
        </lib>
    </glyph>

    Example from "B" in SourceCodePro-Regular
    <key><com.adobe.type.autohint><key>
    <dict>
        <key>id</key>
        <string>64bf4987f05ced2a50195f971cd924984047eb1d79c8c43e6a0054f59cc85dea23
        a49deb20946a4ea84840534363f7a13cca31a81b1e7e33c832185173369086</string>
        <key>hintSetList</key>
        <array>
            <dict>
                <key>pointTag</key>
                <string>hintSet0000</string>
                <key>stems</key>
                <array>
                    <string>hstem 338 28</string>
                    <string>hstem 632 28</string>
                    <string>hstem 100 32</string>
                    <string>hstem 496 32</string>
                </array>
            </dict>
            <dict>
                <key>pointTag</key>
                <string>hintSet0005</string>
                <key>stems</key>
                <array>
                    <string>hstem 0 28</string>
                    <string>hstem 338 28</string>
                    <string>hstem 632 28</string>
                    <string>hstem 100 32</string>
                    <string>hstem 454 32</string>
                    <string>hstem 496 32</string>
                </array>
            </dict>
            <dict>
                <key>pointTag</key>
                <string>hintSet0016</string>
                <key>stems</key>
                <array>
                    <string>hstem 0 28</string>
                    <string>hstem 338 28</string>
                    <string>hstem 632 28</string>
                    <string>hstem 100 32</string>
                    <string>hstem 496 32</string>
                </array>
            </dict>
        </array>
    <dict>
    """
    hint_sets = []
    stems: "list[HintTuple]" = []
    hint_set: HintSet = HintSet(pointTag="0", stems=stems)
    if mmglyph.hintmasks:
        for mask in mmglyph.hintmasks:
            for direction in DIRECTIONS:
                dd, hint_index = mask
                if direction == dd:
                    master_direction_hints = master_hints[direction]
                    if hint_index < len(master_direction_hints):
                        hint: "HintTuple" = master_direction_hints[hint_index]
                        hint_set["stems"].append(hint)
                    else:
                        logger.debug(
                            f"Hint mask '{direction}' with index {hint_index} found in "
                            f"glyph {glyph.name}, but hint list is empty."
                        )
            if mask[0] == "r":
                hint_set["pointTag"] = mmglyph.get_point_label(
                    index=int(hint_set["pointTag"]),
                    code="PSHintReplacement",
                    start_count=0,
                )
                hint_sets.append(hint_set)
                node_index = mask[1]
                # FIXME: What do negative values mean?
                if node_index < 0:
                    node_index = abs(node_index) - 1
                stems = []
                hint_set = HintSet(pointTag=str(node_index), stems=stems)

        if hint_set["stems"]:
            # Append the last hint set
            hint_set["pointTag"] = mmglyph.get_point_label(
                index=int(hint_set["pointTag"]),
                code="PSHintReplacement",
                start_count=0,
            )
            hint_sets.append(hint_set)
    else:
        # Only one hint set, always make a hint set with first point
        for direction in DIRECTIONS:
            for hint in master_hints[direction]:
                hint_set["stems"].append(hint)
        if hint_set["stems"]:
            hint_set["pointTag"] = mmglyph.get_point_label(
                index=int(hint_set["pointTag"]),
                code="PSHintReplacement",
                start_count=0,
            )
            hint_sets = [hint_set]

    # Reformat stems from sortable tuples to str required by UFO spec
    ufo_hint_sets = {}
    for hint_set in hint_sets:
        point_tag = hint_set["pointTag"]
        ufo_hint_set = UfoHintSet(
            pointTag=point_tag,
            stems=[f"{h[0]} {h[1]} {h[2]}" for h in sorted(set(hint_set["stems"]))],
        )
        if ufo_hint_set["stems"]:
            # Only add if the set has stems
            if ufo_hint_set["pointTag"] in ufo_hint_sets:
                if ufo_hint_sets[point_tag] != ufo_hint_set:
                    # Warn about duplicate hint sets per point, but use the last one
                    logger.warning(
                        f"Duplicate hint sets for point '{point_tag}' "
                        f"in glyph '{glyph.name}':"
                    )
                    logger.warning(f"    Old: {ufo_hint_sets[point_tag]}")
                    logger.warning(f"    New: {ufo_hint_set}")
                    logger.warning("    Using the new hint set.")
            ufo_hint_sets[point_tag] = ufo_hint_set

    if ufo_hint_sets:
        if not hasattr(glyph, "lib"):
            glyph.lib = {}
        glyph.lib[PS_GLYPH_LIB_KEY] = {
            "formatVersion": "1",
            # "id": "FIXME",
            "hintSetList": list(ufo_hint_sets.values()),
            # "flexList": [],
        }


def get_master_hints(
    mmglyph: "VfbToUfoGlyph", master_index=0
) -> "dict[str, list[HintTuple]]":
    hints: "dict[str, list[HintTuple]]" = {"h": [], "v": []}

    # Hints
    for direction in DIRECTIONS:
        dh = mmglyph.mm_hints[direction]
        for mm_hints in dh:
            hint = mm_hints[master_index]
            hint = normalize_hint_dict(hint, f"{direction}stem")
            hints[direction].append(hint)

    # Links
    if not mmglyph.links:
        return hints

    # Convert links to hints
    for i, axis in enumerate("xy"):
        direction_links = mmglyph.links[axis]
        for link in direction_links:
            isrc, itgt = link  # indices of source and target node
            src = mmglyph.mm_nodes[isrc]
            src_pos = src["points"][master_index][0][i]
            pos = src_pos
            if itgt == -1:  # Bottom ghost
                width = -21
                pos = src_pos - width
            elif itgt == -2:  # Top ghost
                width = -20
            else:
                tgt = mmglyph.mm_nodes[itgt]
                tgt_pos = tgt["points"][master_index][0][i]
                width = tgt_pos - src_pos
                # pos = min(src_pos, tgt_pos)

            d = "v" if axis == "x" else "h"
            # Don't normalize those values, the above code already did that
            hint = (f"{d}stem", pos, width)
            hints[d].append(hint)

    return hints


def update_adobe_hinting(data) -> UfoHintingV2:
    # Convert Adobe hinting data from v1 to v2.
    # https://github.com/adobe-type-tools/psautohint/blob/master/python/psautohint/ufoFont.py
    try:
        # Data may be base64-encoded
        data = data.decode()
    except AttributeError:
        pass
    if not isinstance(data, str):
        # V1 data is stored as str, so if it is not a str, we have nothing to do
        return data

    v2 = UfoHintingV2()
    root = elementTree.fromstring(data)
    hintset: UfoHintSet | None = None
    hintSetList: list[UfoHintSet] = []
    for el in root.iter():
        if el.tag == "hintSetList":
            hintSetList = []
        elif el.tag == "hintset":
            if hintset:
                hintSetList.append(hintset)
            hintset = {
                "pointTag": el.attrib["pointTag"],
                "stems": [],
            }
        elif el.tag in ("hstem", "vstem"):
            if hintset is not None:
                tag, pos, width = normalize_hint(
                    (el.tag, int(el.attrib["pos"]), int(el.attrib["width"]))
                )
                hintset["stems"].append(f"{tag} {pos} {width}")
    if hintset:
        hintSetList.append(hintset)
    if hintSetList:
        v2["hintSetList"] = hintSetList

    # Remove empty entries
    empty = [k for k in v2 if not v2[k]]
    for k in empty:
        del v2[k]
    return v2



import logging
from functools import cached_property
from typing import TYPE_CHECKING

from fontTools.pens.pointPen import AbstractPointPen, SegmentToPointPen

from vfbLib import DIRECTIONS
from vfbLib.templates.glyph import get_empty_glyph
from vfbLib.ufo.glyph import VfbToUfoGlyph
from vfbLib.ufo.paths import UfoMasterGlyph
from vfbLib.ufo.pshints import normalize_hint_dict
from vfbLib.vfb.pens import VfbGlyphPointPen

if TYPE_CHECKING:
    from fontTools.pens.basePen import AbstractPen

    from vfbLib.typing import HintTuple
    from vfbLib.vfb.entry import VfbEntry
    from vfbLib.vfb.vfb import Vfb, VfbMaster


logger = logging.getLogger(__name__)


class VfbGlyph:
    def __init__(self, entry: "VfbEntry", parent: "Vfb | VfbMaster") -> None:
        self.entry = entry

        # Related entries that we need when generating fonts
        self.links_entry: "VfbEntry | None" = None
        self.ps_hinting_options: "VfbEntry | None" = None

        self._parent = parent
        self._glyph: UfoMasterGlyph | None = None
        self.master_index = 0
        self.empty()

    # UFO/cu2qu compatibility

    def __len__(self) -> int:
        """
        Return the number of paths in the glyph.

        Returns:
            int: The number of paths, or contours.
        """
        if not self.entry.data["nodes"]:
            return 0

        node_types = [node["type"] for node in self.entry.data["nodes"]]
        return node_types.count("move")

    @cached_property
    def name(self) -> str | None:
        if self._glyph is None:
            return None

        return self._glyph.name

    def clearContours(self):
        try:
            del self.entry.data["hints"]
        except KeyError:
            pass
        self.entry.data["nodes"] = []

    # Native methods

    def decompile(self) -> str:
        """
        Decompile the Glyph entry and return the glyph name.
        """
        self.entry.decompile()
        if self.entry.data is None:
            raise ValueError

        return self.entry.data["name"]

    def empty(self) -> None:
        self.entry.data = get_empty_glyph(self._parent.num_masters)

    def _copy_to_ufo_glyph(self) -> None:
        """
        Copy minimal data to the VfbToUfoGlyph. Only data that is necessary for the pen
        methods is copied.
        """
        if self.entry.data is None:
            raise ValueError

        _mm_glyph = VfbToUfoGlyph()
        _mm_glyph.name = self.entry.data["name"]

        if "components" in self.entry.data:
            _mm_glyph.mm_components = self.entry.data["components"]

        if "metrics" in self.entry.data:
            _mm_glyph.mm_metrics = self.entry.data["metrics"]

        if "nodes" in self.entry.data:
            _mm_glyph.mm_nodes = self.entry.data["nodes"]

        # TODO: Support point names used by TT hinting
        # if "tth" in self.entry.data:
        #     _mm_glyph.tth_commands = self.entry.data["tth"]

        self._glyph = UfoMasterGlyph(
            mm_glyph=_mm_glyph,
            glyph_order=self._parent.glyph_order,
            master_index=self._parent.master_index,
        )
        self._glyph.build()

    def resolve_hint_sets(self) -> "dict[int, list[HintTuple]]":
        hint_sets = {}
        hints = self.resolve_hints()
        hints_entry = self.entry.data.get("hints", {})
        hintmasks = hints_entry.get("hintmasks", [])
        if hintmasks:
            set_index = 0
            set_hints = []
            for kind, index in hintmasks:
                if kind == "r":
                    if set_hints:
                        if set_index in hint_sets:
                            logger.warning(f"Duplicate hint sets for node {set_index}")
                        hint_sets[set_index] = set_hints
                    set_index = index
                    set_hints = []
                else:
                    set_hints.append(hints[kind][index])
            if set_hints:
                # TODO: Filter out any consecutive identical hintmasks
                hint_sets[set_index] = set_hints
        else:
            # All in one hint set
            hint_sets[0] = hints["h"] + hints["v"]
        return hint_sets

    def resolve_hints(self) -> "dict[str, list[HintTuple]]":
        hints: "dict[str, list[HintTuple]]" = {"h": [], "v": []}

        mm_hints = self.entry.data.get("hints", {"h": [], "v": []})

        # Hints
        for direction in DIRECTIONS:
            direction_hints = mm_hints[direction]
            for mm_hint in direction_hints:
                hint = mm_hint[self.master_index]
                hint = normalize_hint_dict(hint, f"{direction}stem")
                hints[direction].append(hint)

        # Links
        links = self.resolve_links()
        for direction in DIRECTIONS:
            hints[direction].extend(links[direction])

        return hints

    def resolve_links(self) -> "dict[str, list[HintTuple]]":
        # Convert the links to hints

        hints = {"h": [], "v": []}

        if self.links_entry is None:
            return hints

        links = self.links_entry.data
        assert isinstance(links, dict)
        if not (links["x"] or links["y"]):
            return hints

        mm_nodes = self.entry.data.get("nodes", [])

        # Do the actual conversion
        for i, axis in enumerate("xy"):
            direction_links = links[axis]
            for link in direction_links:
                isrc, itgt = link  # indices of source and target node
                src = mm_nodes[isrc]
                src_pos = src["points"][self.master_index][0][i]
                pos = src_pos
                if itgt == -1:  # Bottom ghost
                    width = -21
                    pos = src_pos - width
                elif itgt == -2:  # Top ghost
                    width = -20
                else:
                    tgt = mm_nodes[itgt]
                    tgt_pos = tgt["points"][self.master_index][0][i]
                    width = tgt_pos - src_pos

                d = "v" if axis == "x" else "h"
                # Don't normalize those values, the above code already did that
                hint = (f"{d}stem", pos, width)
                hints[d].append(hint)
        return hints

    def draw(self, pen: "AbstractPen") -> None:
        """
        Draw the VFB glyph onto a segment pen. Uses the object's `master_index` property
        to determine which master is drawn.
        """
        if self.entry.data is None:
            raise ValueError

        glyph = self.entry.data

        path_is_open = False
        in_path = False
        in_qcurve = False
        qcurve = []

        for n in glyph.get("nodes", []):
            node_type = n.get("type")
            flags = n.get("flags", 0)
            pt = n["points"][self.master_index][0]
            if node_type == "move":
                if in_qcurve:
                    qcurve.append(pt)
                    pen.qCurveTo(*qcurve)
                if in_path:
                    if path_is_open:
                        pen.endPath()
                    else:
                        pen.closePath()
                    in_path = False
                in_qcurve = False
                qcurve = []
                pen.moveTo(pt)
                in_path = True
                path_is_open = bool(flags & 8)

            elif node_type == "line":
                if in_qcurve:
                    qcurve.append(pt)
                    pen.qCurveTo(*qcurve)
                    in_qcurve = False
                    qcurve = []
                else:
                    pen.lineTo(pt)
            elif node_type == "curve":
                # FIXME: What about mixed curve/qcurve?
                pt, c1, c2 = n["points"][self.master_index]
                pen.curveTo(c1, c2, pt)
            elif node_type == "qcurve":
                if not in_qcurve:
                    in_qcurve = True
                qcurve.append(pt)
            else:
                raise ValueError(f"Unknown node type: '{node_type}'")
        if in_path:
            if path_is_open:
                pen.endPath()
            else:
                pen.closePath()

        for c in glyph.get("components", []):
            baseGlyphName = self._parent.glyph_order[c["gid"]]
            transformation = (
                c["scaleX"][self.master_index],
                0,
                0,
                c["scaleY"][self.master_index],
                c["offsetX"][self.master_index],
                c["offsetY"][self.master_index],
            )
            pen.addComponent(baseGlyphName, transformation)

    def drawPoints(self, pen: AbstractPointPen) -> None:
        """
        Draw the VFB glyph onto a point pen.
        """
        if self.entry.data is None:
            raise ValueError

        if self._glyph is None:
            self._copy_to_ufo_glyph()

        if self._glyph is None:
            raise ValueError

        return self._glyph.drawPoints(pen)

    def drawPointsDirect(self, pen: AbstractPointPen) -> None:
        """
        Draw the VFB glyph onto a segment pen. Uses the object's `master_index` property
        to determine which master is drawn.
        """
        if self.entry.data is None:
            raise ValueError

        glyph = self.entry.data

        in_path = False
        in_qcurve = False

        for n in glyph.get("nodes", []):
            node_type = n.get("type")
            if node_type == "move":
                if in_qcurve:
                    pen.addPoint(n["points"][self.master_index][0], "qcurve")
                    in_qcurve = False
                if in_path:
                    pen.endPath()
                pen.beginPath()
                in_path = True
                pen.addPoint(n["points"][self.master_index][0], "move")
            elif node_type == "line":
                if in_qcurve:
                    pen.addPoint(n["points"][self.master_index][0], "qcurve")
                    in_qcurve = False
                else:
                    pen.addPoint(n["points"][self.master_index][0], "line")
            elif node_type == "curve":
                pt3, pt1, pt2 = n["points"][self.master_index]
                pen.addPoint(pt1, None)
                pen.addPoint(pt2, None)
                pen.addPoint(pt3, "curve")
            elif node_type == "qcurve":
                if not in_qcurve:
                    in_qcurve = True
                pen.addPoint(n["points"][self.master_index][0], None)
            else:
                raise ValueError(f"Unknown node type: '{node_type}'")
        if in_path:
            pen.endPath()

        for c in glyph.get("components", []):
            baseGlyphName = self._parent.glyph_order[c["gid"]]
            transformation = (
                c["scaleX"][self.master_index],
                0,
                0,
                c["scaleY"][self.master_index],
                c["offsetX"][self.master_index],
                c["offsetY"][self.master_index],
            )
            pen.addComponent(baseGlyphName, transformation)

    def getPen(self) -> "AbstractPen":
        """
        Return a segment pen to draw into the VFB glyph.
        """
        # TODO: Test
        return SegmentToPointPen(self.getPointPen(), guessSmooth=True)

    def getPointPen(self) -> VfbGlyphPointPen:
        """
        Return a point pen to draw into the VFB glyph.
        """
        # FIXME: Add test!!!
        if self.entry.data is None:
            if self.entry.data is None:
                # Make an empty glyph
                self.empty()
            else:
                self.decompile()
        return VfbGlyphPointPen(self, self._parent)


class VfbGlyphMaster:
    def __init__(self, glyph: VfbGlyph, master_index: int = 0):
        self.glyph = glyph
        self.entry = glyph.entry
        self.master_index = master_index
        self._glyph = None

    def __len__(self) -> int:
        """
        Return the number of paths in the glyph.

        Returns:
            int: The number of paths, or contours.
        """
        if not self.entry.data["nodes"]:
            return 0

        node_types = [node["type"] for node in self.entry.data["nodes"]]
        return node_types.count("move")

    @cached_property
    def name(self) -> str | None:
        if self._glyph is None:
            return None

        return self._glyph.name

    def _copy_to_ufo_glyph(self, master_index):
        """
        Copy minimal data to the VfbToUfoGlyph. Only data that is necessary for the pen
        methods is copied.
        """
        if self.entry.data is None:
            raise ValueError

        _mm_glyph = VfbToUfoGlyph()
        _mm_glyph.name = self.entry.data["name"]

        if "components" in self.entry.data:
            _mm_glyph.mm_components = self.entry.data["components"]

        if "metrics" in self.entry.data:
            _mm_glyph.mm_metrics = self.entry.data["metrics"]

        if "nodes" in self.entry.data:
            _mm_glyph.mm_nodes = self.entry.data["nodes"]

        # TODO: Support point names used by TT hinting
        # if "tth" in self.entry.data:
        #     _mm_glyph.tth_commands = self.entry.data["tth"]

        self._glyph = UfoMasterGlyph(
            mm_glyph=_mm_glyph,
            glyph_order=self.glyph._parent.glyph_order,
            master_index=master_index,
        )
        self._glyph.build()

    def clearContours(self):
        if self.master_index == 0:
            target = self.entry.data
        else:
            if self.entry.temp_masters is None:
                self.entry.temp_masters = [
                    {} for _ in range(self.glyph._parent.num_masters)
                ]
            target = self.entry.temp_masters[self.master_index]
        try:
            del target["hints"]
        except KeyError:
            pass
        target["nodes"] = []

    def drawPoints(self, pen: AbstractPointPen) -> None:
        """
        Draw the VFB glyph onto a point pen.
        """
        if self.entry.data is None:
            raise ValueError

        if self._glyph is None:
            self._copy_to_ufo_glyph(self.master_index)

        if self._glyph is None:
            raise ValueError

        return self._glyph.drawPoints(pen)

    def getPen(self) -> "AbstractPen":
        """
        Return a segment pen to draw into the VFB glyph.
        """
        # TODO: Test
        return SegmentToPointPen(self.getPointPen(), guessSmooth=True)

    def getPointPen(self) -> VfbGlyphPointPen:
        """
        Return a point pen to draw into the VFB glyph.
        """
        # FIXME: Add test!!!
        if self.entry.data is None:
            if self.entry.data is None:
                # Make an empty glyph
                self.glyph.empty()
            else:
                self.entry.decompile()
        return VfbGlyphPointPen(self, self.glyph._parent)



import logging
from io import BytesIO
from struct import pack
from typing import TYPE_CHECKING, Any

from fontTools.ttLib.tables.ttProgram import Program

from vfbLib import DIRECTIONS, GLYPH_CONSTANT, gdef_class_names
from vfbLib.compilers.base import BaseCompiler, StreamWriter
from vfbLib.compilers.guides import GuidesCompiler
from vfbLib.parsers.glyph import PathCommand
from vfbLib.truetype import TT_COMMAND_CONSTANTS, TT_COMMANDS

if TYPE_CHECKING:
    from vfbLib.typing import (
        AnchorPropertiesDict,
        GdefDict,
        GlyphData,
        LinkDict,
        MaskData,
        MMAnchorDict,
    )


logger = logging.getLogger(__name__)


class GlyphAnchorsCompiler(BaseCompiler):
    def _compile(self, data: "list[MMAnchorDict]") -> None:
        assert self.vfb is not None
        self.write_value(len(data), signed=False)
        self.write_value(self.vfb.num_masters, signed=False)
        for anchor in data:
            for i in range(self.vfb.num_masters):
                self.write_value(anchor["x"][i])
                self.write_value(anchor["y"][i])


class GlyphAnchorsSuppCompiler(BaseCompiler):
    def _compile(self, data: "list[AnchorPropertiesDict]") -> None:
        self.write_value(len(data), signed=False)
        for anchor in data:
            self.write_value(anchor.get("hue", 0), signed=False)
            self.write_value(anchor.get("data", 0), signed=False)


class GlyphGDEFCompiler(BaseCompiler):
    def _compile(self, data: "GdefDict") -> None:
        c = data.get("glyph_class", "unassigned")
        if c is None:
            gdef_class = 0
        else:
            gdef_class = gdef_class_names.index(c)
        self.write_value(gdef_class, signed=False)

        anchors = data.get("anchors", [])
        self.write_value(len(anchors), signed=False)
        for anchor in anchors:
            self.write_str_with_len(anchor.get("name", ""))
            self.write_value(anchor["x"])
            self.write_value(-1)
            self.write_value(anchor["y"])
            self.write_value(-1)

        carets = data.get("carets", [])
        self.write_value(len(carets), signed=False)
        for pos, xxx in carets:
            self.write_value(pos)
            self.write_value(xxx)

        ot_classes = data.get("ot_classes", [])
        self.write_value(len(ot_classes), signed=False)
        for value in ot_classes:
            self.write_value(value)


class GlyphCompiler(BaseCompiler):
    @classmethod
    def merge(cls, masters_data: list[Any], data: Any) -> None:
        num_masters = len(masters_data)
        if num_masters < 2:
            return

        for m in range(1, num_masters):
            master_data = masters_data[m]
            # See if there is any data to merge
            if "nodes" in master_data:
                assert "nodes" in data
                for i, tgt in enumerate(data["nodes"]):
                    src = master_data["nodes"][i]
                    for key in ("type", "flags"):
                        assert src[key] == tgt[key]
                    tgt["points"][m] = src["points"][m]

            if "components" in master_data:
                assert "components" in data
                for i, tgt in enumerate(data["components"]):
                    src = master_data["components"][i]
                    for key in ("gid",):
                        assert src[key] == tgt[key]
                    tgt["offsetX"][m] = src["offsetX"][m]
                    tgt["offsetY"][m] = src["offsetY"][m]
                    tgt["scaleX"][m] = src["scaleX"][m]
                    tgt["scaleY"][m] = src["scaleY"][m]

    def _compile(self, data: dict[str, Any]) -> None:
        # Constants
        self.write_bytes(pack("<4B", *GLYPH_CONSTANT))
        self.num_masters = data["num_masters"]

        self._compile_glyph_name(data)
        self.compile_outlines(data)
        self._compile_metrics(data)
        self._compile_hints(data)
        self._compile_guides(data)
        self._compile_components(data)
        self._compile_kerning(data)
        self._compile_binary(data)
        self._compile_instructions(data)
        self.write_uint8(15)  # End of glyph

    def _compile_binary(self, data):
        # Imported binary data 8-)
        if not (imported := data.get("imported")):  # noqa: F841
            return

        self.write_uint8(9)

        self.write_uint8(0x29)  # Metrics
        self.write_value(imported["width"], signed=False)
        self.write_value(imported["lsb"])
        self.write_value(imported["unknown1"])
        self.write_value(imported["unknown2"])
        self.write_value(imported["unknown3"])
        for value in imported["bbox"]:
            self.write_value(value)

        # Outlines

        self.write_uint8(0x2A)
        num_contours = imported["num_contours"]  # -1 for composite, so we don't count
        self.write_value(num_contours)
        endpoints = imported.get("endpoints", [])
        if num_contours >= 0 and len(endpoints) != num_contours:
            logger.warning(
                f"Imported binary glyph: Number of contours ({num_contours}) doesn't "
                f"match length of endpoints array ({len(endpoints)})"
            )
        for endpoint in endpoints:
            self.write_value(endpoint, signed=False)
        nodes = imported.get("nodes", [])
        self.write_value(len(nodes), signed=False)  # num_nodes
        x0 = 0
        y0 = 0
        for node in nodes:
            x, y = node["point"]
            self.write_value(x - x0)
            self.write_value(y - y0)
            x0 = x
            y0 = y
            self.write_uint8(node["flags"])

        # TrueType instructions

        instructions = imported.get("instructions")
        self.write_uint8(0x2B)
        if instructions:
            p = Program()
            p.fromAssembly(instructions)
            bytecode = p.getBytecode()
            self.write_value(len(bytecode), signed=False)  # num_bytes
            self.write_bytes(bytecode)
        else:
            self.write_value(0, signed=False)

        # HDMX data

        hdmx = imported.get("hdmx", [])
        self.write_uint8(0x2C)
        self.write_value(len(hdmx), signed=False)
        for value in hdmx:
            self.write_uint8(value)

        self.write_uint8(0x28)  # end

    def _compile_components(self, data):
        # Components
        if not (components := data.get("components")):
            return

        self.write_uint8(5)
        self.write_value(len(components))
        for component in components:
            self.write_value(component["gid"])
            for i in range(self.num_masters):
                self.write_value(component["offsetX"][i])
                self.write_value(component["offsetY"][i])
                self.write_double(component["scaleX"][i])
                self.write_double(component["scaleY"][i])

    def _compile_glyph_name(self, data):
        # Glyph name
        if not (name := data.get("name")):
            return

        glyph_name = name.encode("cp1252")
        glyph_name_length = len(glyph_name)
        self.write_uint8(1)
        self.write_value(glyph_name_length)
        self.write_bytes(glyph_name)
        logger.debug(f"Compiling glyph '{name}'")

    def _compile_guides(self, data):
        # Guidelines
        if not (guides := data.get("guides")):
            # TODO: Do we always need to write the guides data?
            # guides = MMGuidesDict(
            #     h=[[] for _ in range(self.num_masters)],
            #     v=[[] for _ in range(self.num_masters)],
            # )
            return

        self.write_uint8(4)
        gc = GuidesCompiler()
        gc.stream = self.stream
        gc.vfb = self.vfb
        gc._compile(guides)

    def _compile_hints(self, data):
        # PostScript hints
        # To minimize diffs, we always write out hint, but it is not necessary
        hints = data.get("hints", {})
        # We could skip empty hinting:
        if not (hints := data.get("hints")):
            return

        self.write_uint8(3)
        for direction in DIRECTIONS:
            if direction_hints := hints.get(direction):
                self.write_value(len(direction_hints))
                for mm_hint in direction_hints:
                    for i in range(self.num_masters):
                        hint = mm_hint[i]
                        self.write_value(hint["pos"])
                        self.write_value(hint["width"])
            else:
                self.write_value(0)

        if not (hintmasks := hints.get("hintmasks")):  # noqa: F841
            self.write_value(0)
            return

        self.write_value(len(hintmasks))
        for k, v in hintmasks:
            key = {
                "h": 0x01,
                "v": 0x02,
                "r": 0xFF,
            }[k]
            self.write_uint8(key)
            self.write_value(v)

    def _compile_instructions(self, data):
        # TrueType instructions
        if not (tth := data.get("tth")):
            # self.write_uint8(0x0A)
            # self.write_value(0)
            return

        self.write_uint8(0x0A)
        instructions = InstructionsCompiler().compile(tth)
        self.write_value(len(instructions))
        self.stream.write(instructions)

    def _compile_kerning(self, data):
        # Kerning
        if not (kerning := data.get("kerning")):
            return

        self.write_uint8(6)
        self.write_value(len(kerning))
        for gid, values in kerning.items():
            self.write_value(gid)
            for value in values:
                self.write_value(value)

    def _compile_metrics(self, data):
        # Metrics
        if not (metrics := data.get("metrics")):
            return

        self.write_uint8(2)
        for i in range(self.num_masters):
            x, y = metrics[i]
            self.write_value(x)
            self.write_value(y)

    def compile_outlines(self, data, write_key=True):
        # Outlines
        # A minimal outlines structure is always written:
        if write_key:
            self.write_uint8(8)
        self.write_value(self.num_masters)

        if not (nodes := data.get("nodes")):
            # 0 nodes with 0 values
            self.write_value(0)
            self.write_value(0)
            return

        outlines, num_values = OutlinesCompiler().compile(nodes, self.num_masters)
        self.write_value(num_values)
        self.stream.write(outlines)


class GlyphOriginCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        self.write_int16(data["x"])
        self.write_int16(data["y"])


class GlyphSketchCompiler(BaseCompiler):
    def _compile(self, data: list[tuple[int, int, int]]) -> None:
        self.write_value(len(data), signed=False)
        for a, b, c in data:
            self.write_value(a)
            self.write_value(b)
            self.write_value(c)


class GlyphUnicodesCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        for value in data:
            self.write_uint16(value)


class GlyphUnicodesSuppCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        for value in data:
            self.write_uint32(value)


class InstructionsCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        self.write_value(len(data))
        for cmd in data:
            command_id = TT_COMMAND_CONSTANTS[cmd["cmd"]]
            self.write_uint8(command_id)
            params = cmd["params"]
            for param_name in TT_COMMANDS[command_id]["params"]:
                self.write_value(params[param_name])
        for _ in range(3):
            self.write_value(0)


class OutlinesCompiler(StreamWriter):
    def compile(self, data: Any, num_masters: int) -> tuple[bytes, int]:
        self.num_masters = num_masters
        self.stream = BytesIO()
        num_values = self._compile(data)
        return self.stream.getvalue(), num_values

    def _compile(self, data: Any) -> int:
        self.write_value(len(data))  # Number of nodes, may be 0
        num_values = 0
        # Reference coordinates to build relative values for each master
        ref_coords = [[0, 0] for _ in range(self.num_masters)]
        for node in data:
            type_flags = node.get("flags", 0) * 16 + PathCommand[node["type"]].value
            self.write_uint8(type_flags)
            num_values += 1
            for j in range(len(node["points"][0])):
                for i in range(self.num_masters):
                    x, y = node["points"][i][j]
                    refx, refy = ref_coords[i]
                    # Coordinates are written relatively to the previous coords
                    self.write_value(x - refx)
                    self.write_value(y - refy)
                    num_values += 2
                    ref_coords[i] = [x, y]
        return 2 * num_values


class LinksCompiler(BaseCompiler):
    def _compile(self, data: "LinkDict") -> None:
        for direction in ("y", "x"):
            dir_links = data[direction]
            self.write_value(len(dir_links))
            for p0, p1 in dir_links:
                self.write_value(p0)
                self.write_value(p1)


class MaskCompiler(GlyphCompiler):
    def _compile(self, data: "MaskData") -> None:
        weight_vector = data["weight_vector"]
        self.write_value(len(weight_vector))
        for value in weight_vector:
            self.write_value(round(value * 100_000_000))

        self.num_masters = data["num_masters"]
        self.compile_outlines(data, write_key=False)


class GlobalMaskCompiler(GlyphCompiler):
    def _compile(self, data: "GlyphData") -> None:
        self.num_masters = data["num_masters"]
        self.compile_outlines(data, write_key=False)


class MaskMetricsCompiler(BaseCompiler):
    def _compile(self, data: tuple[int, int]) -> None:
        x, y = data
        self.write_int16(x)
        self.write_int16(y)


class MaskMetricsMMCompiler(BaseCompiler):
    def _compile(self, data: list[tuple[int, int]]) -> None:
        for value in data:
            x, y = value
            self.write_value(x)
            self.write_value(y)



from io import BytesIO
from typing import TYPE_CHECKING

from vfbLib.compilers.base import StreamWriter

if TYPE_CHECKING:
    from vfbLib.typing import VfbHeaderDict


class VfbHeaderCompiler(StreamWriter):
    def compile(self, data: "VfbHeaderDict") -> bytes:
        self.stream = BytesIO()
        self._compile(data)
        return self.stream.getvalue()

    def _compile(self, data: "VfbHeaderDict") -> None:
        self.write_uint32(data["signature"])
        self.write_uint8(data["app_version"])
        self.write_uint8(data["file_version"])
        self.write_uint8(data["version_major"])
        self.write_uint8(data["version_minor"])
        # TODO: How is the padding length determined?
        data_offset = 44
        self.write_uint32(data_offset)
        bytes_written = 12
        for _ in range(data_offset - bytes_written):
            self.write_uint8(0)



import logging
from itertools import chain, groupby
from typing import TYPE_CHECKING

from vfbLib.compilers.base import BaseCompiler

if TYPE_CHECKING:
    from vfbLib.typing import BackgroundImageDict, BitmapDataDict, GlyphBitmapDict

logger = logging.getLogger(__name__)


def flush_buffer(
    target: list[list[int]], buf: list[int], max_chunk_len: int = 128
) -> None:
    # The buffer must be split into chunks of max. max_chunk_len (128) values, because
    # its length needs to be written as a signed int8 (len(buf) - 1) which has a maximum
    # value of 127.
    for i in range(0, len(buf), max_chunk_len):
        chunk = buf[i : i + max_chunk_len]  # noqa: E203
        target.append([len(chunk) - 1] + chunk)


class BaseBitmapCompiler(BaseCompiler):
    def _encode_bitmap(self, data: list[list[int]]) -> list[list[int]]:
        # Encode to the run-length-encoded format

        # Returns a list of lists in which the first element is the repetition specifier
        # and the rest of the values are the byte values.
        # A positive number means the following bytes should be used verbatim.
        # E.g.: [[3, 128, 128, 64, 32]]: Use the following 3 + 1 bytes verbatim
        # A negative first number means the following byte should be repeated.
        # The number of repetitions is -n + 1.
        # E.g. [[-2, 0]]: Use the 0 three times

        max_chunk_len = 128

        encoded: list[list[int]] = []
        buf: list[int] = []
        for value, expanded in groupby(chain.from_iterable(data)):
            d = list(expanded)
            count = len(d)
            if count > 2:
                # Repeated value

                # Flush the buffer before handling the repetition.
                flush_buffer(encoded, buf, max_chunk_len)
                buf = []
                # Write the repeated value.
                # It also needs to be split into chunks.
                full_chunks = count // max_chunk_len
                for _ in range(full_chunks):
                    encoded.append([-max_chunk_len + 1, value])
                partial_chunk_length = count % max_chunk_len
                if partial_chunk_length:
                    encoded.append([-partial_chunk_length + 1, value])
            else:
                # Verbatim values

                # The buffer must be written in chunks, but we take care of that when it
                # is flushed (see above).
                buf.extend(d)

        # Flush whatever is left in the buffer
        flush_buffer(encoded, buf, max_chunk_len)
        return encoded

    def _compile_bitmap_data(self, bitmap: "BitmapDataDict") -> None:
        # For the "windows" platform, the bitmap must be inverted.
        data = bitmap["data"]
        if self.vfb is not None and self.vfb.writer_platform == "windows":
            data = [[~b - 0x100 for b in row] for row in data]

        encoded = self._encode_bitmap(data)
        num_values = len(list(chain.from_iterable(encoded)))
        self.write_value(num_values)
        for entry in encoded:
            bytes_spec = entry.pop(0)
            self.write_int8(bytes_spec)
            for b in entry:
                self.write_uint8(b)


class BackgroundBitmapCompiler(BaseBitmapCompiler):
    def _compile(self, data: "BackgroundImageDict") -> None:
        x, y = data["origin"]
        self.write_value(x)
        self.write_value(y)

        w, h = data["size_units"]
        self.write_value(w, signed=False)
        self.write_value(h, signed=False)

        w, h = data["size_pixels"]
        self.write_value(w, signed=False)
        self.write_value(h, signed=False)

        bitmap = data["bitmap"]
        self._compile_bitmap_data(bitmap)


class GlyphBitmapsCompiler(BaseBitmapCompiler):
    def _compile(self, data: "list[GlyphBitmapDict]") -> None:
        self.write_value(len(data), signed=False)
        for d in data:
            self.write_value(d["ppm"], signed=False)
            x, y = d["origin"]
            self.write_value(x)
            self.write_value(y)

            w, h = d["adv"]
            self.write_value(w, signed=False)
            self.write_value(h, signed=False)

            w, h = d["size_pixels"]
            self.write_value(w, signed=False)
            self.write_value(h, signed=False)

            bitmap = d["bitmap"]
            self._compile_bitmap_data(bitmap)



import logging
from collections.abc import Iterable
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any

from vfbLib.enum import F, G, T
from vfbLib.vfb.entry import VfbEntry
from vfbLib.vfb.glyph import VfbGlyph, VfbGlyphMaster
from vfbLib.vfb.header import VfbHeader
from vfbLib.vfb.info import VfbInfo

if TYPE_CHECKING:
    from io import BufferedIOBase

    from vfbLib.typing import VfbDict


logger = logging.getLogger(__name__)


# Convenience objects for vfb access


class VfbMaster:
    """
    Minimal UFO interface for a single master of a multiple master VFB.
    """

    def __init__(self, vfb: "Vfb", master_index: int = 0):
        self.vfb = vfb
        self.master_index = master_index

    def __contains__(self, key: str) -> bool:
        return key in self.vfb

    def __getitem__(self, key: str) -> VfbGlyphMaster:
        return self.vfb.getGlyphMaster(key, self.master_index)

    @property
    def glyph_order(self) -> list[str]:
        return self.vfb.glyph_order

    @property
    def info(self) -> VfbInfo:
        return self.vfb.info

    def items(self) -> Iterable[tuple[str, VfbGlyph]]:
        return self.vfb.items()

    def keys(self) -> Iterable[str]:
        return self.vfb.keys()

    @property
    def num_masters(self) -> int:
        return self.vfb.num_masters

    @property
    def ps_hinting_options(self) -> VfbEntry | None:
        return self.vfb.ps_hinting_options


class Vfb:
    """
    Object to represent the vfb data, with the ability to read and write. You can use
    the Vfb object to access glyphs through dict methods, where the glyph name is the
    key and the glyph object is the value.
    """

    def __init__(
        self,
        vfb_path: Path | None = None,
        timing=True,
        minimal=False,
        drop_keys: set[int] | None = None,
        only_header=False,
        unicode_strings=False,
    ) -> None:
        self.vfb_path = vfb_path
        self.timing = timing
        self.minimal = minimal
        if drop_keys is None:
            self.drop_keys: set[int] = set()
        else:
            self.drop_keys = set(drop_keys)
        self.only_header = only_header
        self.encoding = "utf-8"
        self.force_unicode_strings = unicode_strings

        # We need some minimal API to make pen access work ...
        self._glyphs: dict[str, VfbGlyph] = {}
        self.glyph_order: list[str] = []

        # cu2qu accesses the info and lib ...
        self.info = VfbInfo(vfb=self)
        self.lib: dict[str, Any] = {}

        self.master_index = 0
        self.num_masters: int = 0
        self.ttStemsV_count: int = 0
        self.ttStemsH_count: int = 0
        self.writer_platform: str = "macos"

        self.ps_hinting_options: VfbEntry | None = None

        # Track decompile errors
        self.any_errors = False
        if self.vfb_path:
            self.read()
        else:
            self.header: VfbHeader = VfbHeader()
            self.entries: list[VfbEntry] = []

    def add_entry(self, entry: VfbEntry) -> None:
        """Add an entry to the Vfb. Also sets the parent of the entry to the Vfb.

        Args:
            entry (VfbEntry): The VfbEntry to be added to Vfb.entries.
        """
        entry.vfb = self
        self.entries.append(entry)

    def as_dict(self) -> "VfbDict":
        """
        Return the Vfb structure as Dict, e.g. for saving as JSON. The dict has the keys
        "header" and "entries".
        """
        d: "VfbDict" = {"header": {}, "entries": []}
        if self.header is not None:
            d["header"] = self.header.as_dict()
        if self.entries:
            d["entries"] = [e.as_dict() for e in self.entries]
        return d

    def clear(self) -> None:
        """
        Clear any data that may have been read before.
        """
        self.header = VfbHeader()
        self.entries = []

    def compile(self) -> None:
        """
        Compile the header and all entries in the VFB.
        """
        self.header.compile()
        for e in self.entries:
            e.compile()

    def _decompile_glyphs(self) -> None:
        name = None
        for entry in self.entries:
            match entry.id:
                case G.Glyph:
                    glyph = VfbGlyph(entry, self)
                    name = glyph.decompile()
                    i = 1
                    # Disambiguate duplicate names
                    if name in self._glyphs:
                        logger.error(f"VFB contains duplicate glyph name: {name}")
                        while f"{name}#{i}" in self._glyphs:
                            i += 1
                        name = f"{name}#{i}"
                    self._glyphs[name] = glyph
                    self.glyph_order.append(name)
                case G.Links:
                    # We need to store the links of the glyph for converting links to
                    # PS hints before writing charstrings
                    entry.decompile()
                    if name is None:
                        logger.error("Links entry without preceding glyph entry")
                    else:
                        self._glyphs[name].links_entry = entry
                case G.HintingOptions:
                    # We need this to decide if we should generate hstem3/vstem3 hints
                    entry.decompile()
                    if name is None:
                        logger.error(
                            "Glyph Hinting Options entry without preceding glyph entry"
                        )
                    else:
                        self._glyphs[name].ps_hinting_options = entry
                case F.PostScriptHintingOptions:
                    # We need this to decide if we should generate flex hints
                    entry.decompile()
                    self.ps_hinting_options = entry

    def __contains__(self, key: str) -> bool:
        if not self._glyphs:
            self._decompile_glyphs()
        return key in self._glyphs

    def __getitem__(self, key: str) -> VfbGlyph:
        if not self._glyphs:
            self._decompile_glyphs()
        return self._glyphs[key]

    def getGlyphMaster(self, key: str, master_index: int) -> VfbGlyphMaster:
        return VfbGlyphMaster(self[key], master_index)

    def get_masters(self) -> list[VfbMaster]:
        return [VfbMaster(self, i) for i in range(self.num_masters)]

    def decompile(self) -> None:
        """
        Decompile all entries, except for the ones listed in `drop_keys`.
        """
        self.any_errors = False
        start = time()
        self.header.decompile()
        for entry in self.entries:
            if entry.id in self.drop_keys:
                continue

            entry.decompile()

        end = time()
        if self.timing:
            print(f"Interpreting binary data took {round((end - start) * 1000)} ms.")

    def items(self) -> Iterable[tuple[str, VfbGlyph]]:
        if not self._glyphs:
            self._decompile_glyphs()
        return self._glyphs.items()

    def keys(self) -> Iterable[str]:
        if not self._glyphs:
            self._decompile_glyphs()
        return self._glyphs.keys()

    def read_stream(self, stream: "BufferedIOBase") -> None:
        """
        Lazily read and parse the vfb stream, i.e. parse the header, but only read the
        binary data of other entries.
        """
        start = time()
        self.header = VfbHeader()
        self.header.read(stream)
        if self.only_header:
            return

        entry: VfbEntry | None = None
        while True:
            try:
                entry = VfbEntry(self)
                entry.read(stream)
            except EOFError:
                break

            # Some info needs to be stored for later access, so we decompile a selection
            # of entries right here.

            if entry is not None:
                if entry.id == F.FLVersion:
                    entry.decompile()
                    if entry.data is not None:
                        self.writer_platform = entry.data["platform"]
                        if (
                            self.force_unicode_strings
                            or self.writer_platform == "macos"
                        ):
                            self.encoding = "utf-8"
                        else:
                            self.encoding = "cp1252"

                elif entry.id == F.MasterCount:
                    entry.decompile()
                    if entry.data is not None:
                        if TYPE_CHECKING:
                            assert isinstance(entry.data, int)
                        self.num_masters = entry.data

                elif entry.id == T.TrueTypeStems:
                    entry.decompile()
                    if entry.data is not None:
                        if TYPE_CHECKING:
                            assert isinstance(entry.data, dict)
                        self.ttStemsV_count = len(entry.data.get("ttStemsV", []))
                        self.ttStemsH_count = len(entry.data.get("ttStemsH", []))

                if entry.id not in self.drop_keys:
                    self.entries.append(entry)

            if entry.id == F.BlockFileDataEnd:
                break

        end = time()
        if self.timing:
            print(
                "Source file was successfully read in "
                f"{round((end - start) * 1000)} ms."
            )

    def read(self) -> None:
        """
        Read data from the file at vfb_path, without decompiling.
        """
        self.clear()
        if self.vfb_path is None:
            return

        with open(self.vfb_path, "rb") as vfb:
            self.read_stream(vfb)

    def read_bytes(self, buffer: "BufferedIOBase") -> None:
        """
        Read data from a buffer, without decompiling.

        Args:
            buffer (BufferedIOBase): The buffered reader.
        """
        self.clear()
        self.read_stream(buffer)

    def write(self, out_path: Path) -> None:
        """
        Compile any entries with changes, and write the VFB to out_path.
        """
        with open(out_path, "wb") as vfb:
            self.write_bytes(vfb)

    def write_bytes(self, buffer: "BufferedIOBase") -> None:
        """
        Compile any entries with changes, and write the VFB data to a buffer.

        Args:
            buffer (BufferedIOBase): The buffer.
        """
        if self.header is None:
            raise ValueError

        self.compile()

        assert isinstance(self.header.data, bytes)
        buffer.write(self.header.data)

        for entry in self.entries:
            buffer.write(entry.header)
            if entry.data is not None:
                # There may be entries without data
                assert isinstance(entry.data, bytes)
                buffer.write(entry.data)


def get_vfb_with_masters(num_masters: int) -> Vfb:
    # Minimal Vfb object used for testing
    vfb = Vfb()
    vfb.num_masters = num_masters
    return vfb



from vfbLib.compilers.base import (
    EncodedValueListCompiler,
    EncodedValueListWithCountCompiler,
    GlyphEncodingCompiler,
    HexStringCompiler,
    MappingModeCompiler,
    OpenTypeKerningClassFlagsCompiler,
    OpenTypeMetricsClassFlagsCompiler,
)
from vfbLib.compilers.binary import BinaryTableCompiler
from vfbLib.compilers.bitmap import BackgroundBitmapCompiler, GlyphBitmapsCompiler
from vfbLib.compilers.cmap import CustomCmapCompiler
from vfbLib.compilers.fl3 import MMKernPairCompiler
from vfbLib.compilers.flversion import FLVersionCompiler
from vfbLib.compilers.glyph import (
    GlobalMaskCompiler,
    GlyphAnchorsCompiler,
    GlyphAnchorsSuppCompiler,
    GlyphCompiler,
    GlyphGDEFCompiler,
    GlyphOriginCompiler,
    GlyphSketchCompiler,
    GlyphUnicodesCompiler,
    GlyphUnicodesSuppCompiler,
    LinksCompiler,
    MaskCompiler,
    MaskMetricsCompiler,
    MaskMetricsMMCompiler,
)
from vfbLib.compilers.guides import GuidePropertiesCompiler, GuidesCompiler
from vfbLib.compilers.mm import (
    AnisotropicInterpolationsCompiler,
    AxisMappingsCompiler,
    AxisMappingsCountCompiler,
    MasterLocationCompiler,
    PrimaryInstancesCompiler,
)
from vfbLib.compilers.numeric import (
    DoubleCompiler,
    DoubleListCompiler,
    Int16Compiler,
    IntListCompiler,
    PanoseCompiler,
    SignedInt16Compiler,
    SignedInt32Compiler,
    UnicodeRangesCompiler,
)
from vfbLib.compilers.options import (
    ExportOptionsCompiler,
    FontOptionsCompiler,
)
from vfbLib.compilers.pclt import PcltCompiler
from vfbLib.compilers.ps import (
    PostScriptGlobalHintingOptionsCompiler,
    PostScriptGlyphHintingOptionsCompiler,
    PostScriptInfoCompiler,
)
from vfbLib.compilers.text import (
    NameRecordsCompiler,
    OpenTypeStringCompiler,
    StringCompiler,
    VendorIdCompiler,
)
from vfbLib.compilers.truetype import (
    GaspCompiler,
    TrueTypeInfoCompiler,
    TrueTypeStemPpems1Compiler,
    TrueTypeStemPpems23Compiler,
    TrueTypeStemPpemsCompiler,
    TrueTypeStemsCompiler,
    TrueTypeZoneDeltasCompiler,
    TrueTypeZonesCompiler,
    VdmxCompiler,
)
from vfbLib.enum import F, G, M, T
from vfbLib.parsers.base import (
    BaseParser,
    EncodedValueListParser,
    EncodedValueListWithCountParser,
    GlyphEncodingParser,
    MappingModeParser,
    OpenTypeKerningClassFlagsParser,
    OpenTypeMetricsClassFlagsParser,
)
from vfbLib.parsers.binary import BinaryTableParser
from vfbLib.parsers.bitmap import BackgroundBitmapParser, GlyphBitmapsParser
from vfbLib.parsers.cmap import CustomCmapParser
from vfbLib.parsers.fl3 import MMKernPairParser
from vfbLib.parsers.flversion import FLVersionParser
from vfbLib.parsers.glyph import (
    GlobalMaskParser,
    GlyphAnchorsParser,
    GlyphAnchorsSuppParser,
    GlyphGDEFParser,
    GlyphOriginParser,
    GlyphParser,
    GlyphSketchParser,
    GlyphUnicodeParser,
    GlyphUnicodeSuppParser,
    LinkParser,
    MaskMetricsMMParser,
    MaskMetricsParser,
    MaskParser,
)
from vfbLib.parsers.guides import GlobalGuidesParser, GuidePropertiesParser
from vfbLib.parsers.mm import (
    AnisotropicInterpolationsParser,
    AxisMappingsCountParser,
    AxisMappingsParser,
    MasterLocationParser,
    PrimaryInstancesParser,
)
from vfbLib.parsers.numeric import (
    DoubleListParser,
    DoubleParser,
    Int16Parser,
    IntListParser,
    PanoseParser,
    SignedInt16Parser,
    SignedInt32Parser,
    UnicodeRangesParser,
)
from vfbLib.parsers.options import ExportOptionsParser, FontOptionsParser
from vfbLib.parsers.pclt import PcltParser
from vfbLib.parsers.ps import (
    PostScriptGlobalHintingOptionsParser,
    PostScriptGlyphHintingOptionsParser,
    PostScriptInfoParser,
)
from vfbLib.parsers.text import NameRecordsParser, OpenTypeStringParser, StringParser
from vfbLib.parsers.truetype import (
    GaspParser,
    TrueTypeInfoParser,
    TrueTypeStemPpems1Parser,
    TrueTypeStemPpems23Parser,
    TrueTypeStemPpemsParser,
    TrueTypeStemsParser,
    TrueTypeZoneDeltasParser,
    TrueTypeZonesParser,
    VdmxParser,
)

# fmt: off
parser_classes = {
    # Sorted by appearance in the VFB
    F.BlockFileDataStart: ("Block File Data Start", BaseParser, HexStringCompiler),
    F.BlockFontStart: ("Block Font Start", BaseParser, HexStringCompiler),
    F.FLVersion: ("FL Version", FLVersionParser, FLVersionCompiler),
    F.BlockNamesStart: ("Block Names Start", BaseParser, HexStringCompiler),
    F.EncodingDefault: ("Encoding Default", GlyphEncodingParser, GlyphEncodingCompiler),
    F.Encoding: ("Encoding", GlyphEncodingParser, GlyphEncodingCompiler),
    F.MMEncType: ("MM Encoding Type", Int16Parser, Int16Compiler),
    F.BlockNamesEnd: ("Block Names End", StringParser, HexStringCompiler),
    F.BlockFontInfoStart: ("Block Font Info Start", StringParser, HexStringCompiler),
    F.font_name: ("font_name", StringParser, StringCompiler),
    F.MasterCount: ("Master Count", Int16Parser, Int16Compiler),
    F.weight_vector: ("weight_vector", DoubleListParser, DoubleListCompiler),
    F.unique_id: ("unique_id", SignedInt32Parser, SignedInt32Compiler),
    F.version: ("version", StringParser, StringCompiler),
    F.notice: ("notice", StringParser, StringCompiler),
    F.full_name: ("full_name", StringParser, StringCompiler),
    F.family_name: ("family_name", StringParser, StringCompiler),
    F.pref_family_name: ("pref_family_name", StringParser, StringCompiler),
    F.menu_name: ("menu_name", StringParser, StringCompiler),
    F.apple_name: ("apple_name", StringParser, StringCompiler),
    F.weight: ("weight", StringParser, StringCompiler),
    F.width: ("width", StringParser, StringCompiler),
    F.License: ("License", StringParser, StringCompiler),
    F.LicenseURL: ("License URL", StringParser, StringCompiler),
    F.copyright: ("copyright", StringParser, StringCompiler),
    F.trademark: ("trademark", StringParser, StringCompiler),
    F.designer: ("designer", StringParser, StringCompiler),
    F.designer_url: ("designer_url", StringParser, StringCompiler),
    F.vendor_url: ("vendor_url", StringParser, StringCompiler),
    F.source: ("source", StringParser, StringCompiler),  # manufacturer, "created by"
    F.is_fixed_pitch: ("is_fixed_pitch", Int16Parser, Int16Compiler),
    F.weight_code: ("weight_code", SignedInt16Parser, SignedInt16Compiler),
    F.italic_angle: ("italic_angle", DoubleParser, DoubleCompiler),
    F.slant_angle: ("slant_angle", DoubleParser, DoubleCompiler),
    F.underline_position: ("underline_position", SignedInt16Parser, SignedInt16Compiler),  # noqa: E501
    F.underline_thickness: ("underline_thickness", Int16Parser, Int16Compiler),
    F.ms_charset: ("ms_charset", Int16Parser, Int16Compiler),
    F.panose: ("panose", PanoseParser, PanoseCompiler),
    F.tt_version: ("tt_version", StringParser, StringCompiler),
    F.tt_u_id: ("tt_u_id", StringParser, StringCompiler),
    F.style_name: ("style_name", StringParser, StringCompiler),
    F.pref_style_name: ("pref_style_name", StringParser, StringCompiler),
    F.mac_compatible: ("mac_compatible", StringParser, StringCompiler),
    F.SampleText: ("Sample Text", StringParser, StringCompiler),
    F.vendor: ("vendor", StringParser, VendorIdCompiler),
    F.xuid: ("xuid", IntListParser, IntListCompiler),
    F.xuid_num: ("xuid_num", Int16Parser, Int16Compiler),
    F.year: ("year", Int16Parser, Int16Compiler),
    F.version_major: ("version_major", Int16Parser, Int16Compiler),
    F.version_minor: ("version_minor", Int16Parser, Int16Compiler),
    F.upm: ("upm", Int16Parser, Int16Compiler),
    F.fond_id: ("fond_id", Int16Parser, Int16Compiler),
    F.PostScriptHintingOptions: ("PostScript Hinting Options", PostScriptGlobalHintingOptionsParser, PostScriptGlobalHintingOptionsCompiler),  # noqa: E501
    F.Collection: ("Collection", EncodedValueListWithCountParser, EncodedValueListWithCountCompiler),  # noqa: E501
    F.blue_values_num: ("blue_values_num", Int16Parser, Int16Compiler),
    F.other_blues_num: ("other_blues_num", Int16Parser, Int16Compiler),
    F.family_blues_num: ("family_blues_num", Int16Parser, Int16Compiler),
    F.family_other_blues_num: ("family_other_blues_num", Int16Parser, Int16Compiler),
    F.stem_snap_h_num: ("stem_snap_h_num", Int16Parser, Int16Compiler),
    F.stem_snap_v_num: ("stem_snap_v_num", Int16Parser, Int16Compiler),
    F.font_style: ("font_style", Int16Parser, Int16Compiler),  # OS/2.fsSelection
    F.pcl_id: ("pcl_id", Int16Parser, Int16Compiler),
    F.vp_id: ("vp_id", Int16Parser, Int16Compiler),
    F.ms_id: ("ms_id", Int16Parser, Int16Compiler),
    F.pcl_chars_set: ("pcl_chars_set", StringParser, StringCompiler),

    T.cvt: ("cvt", BaseParser, HexStringCompiler),  # Binary cvt Table
    T.prep: ("prep", BaseParser, HexStringCompiler),  # Binary prep Table
    T.fpgm: ("fpgm", BaseParser, HexStringCompiler),  # Binary fpgm Table
    T.gasp: ("gasp", GaspParser, GaspCompiler),
    F.ttinfo: ("ttinfo", TrueTypeInfoParser, TrueTypeInfoCompiler),
    T.vdmx: ("vdmx", VdmxParser, VdmxCompiler),
    T.hhea_line_gap: ("hhea_line_gap", Int16Parser, Int16Compiler),
    T.hhea_ascender: ("hhea_ascender", SignedInt16Parser, SignedInt16Compiler),
    T.hhea_descender: ("hhea_descender", SignedInt16Parser, SignedInt16Compiler),
    # hstem_data and vstem_data, goes to font.ttinfo:
    T.TrueTypeStemPPEMs2And3: ("TrueType Stem PPEMs 2 And 3", TrueTypeStemPpems23Parser, TrueTypeStemPpems23Compiler),  # noqa: E501
    T.TrueTypeStemPPEMs: ("TrueType Stem PPEMs", TrueTypeStemPpemsParser, TrueTypeStemPpemsCompiler),  # noqa: E501
    # Probably in font.ttinfo, but not accessible through API:
    T.TrueTypeStems: ("TrueType Stems", TrueTypeStemsParser, TrueTypeStemsCompiler),
    T.TrueTypeStemPPEMs1: ("TrueType Stem PPEMs 1", TrueTypeStemPpems1Parser, TrueTypeStemPpems1Compiler),  # noqa: E501
    # Probably in font.ttinfo, but not accessible through API:
    T.TrueTypeZones: ("TrueType Zones", TrueTypeZonesParser, TrueTypeZonesCompiler),

    # Goes to font:
    F.unicoderanges: ("unicoderanges", UnicodeRangesParser, UnicodeRangesCompiler),

    # Probably in font.ttinfo, but not accessible through API:
    T.stemsnaplimit: ("stemsnaplimit", Int16Parser, Int16Compiler),  # Pixel Snap
    T.zoneppm: ("zoneppm", Int16Parser, Int16Compiler),  # Zone Stop PPEM
    T.codeppm: ("codeppm", Int16Parser, Int16Compiler),  # Code Stop PPEM
    T.dropoutppm: ("dropoutppm", Int16Parser, Int16Compiler),  # Dropout control PPEM
    T.MeasurementLine: ("Measurement Line", SignedInt16Parser, SignedInt16Compiler),
    T.TrueTypeZoneDeltas: ("TrueType Zone Deltas", TrueTypeZoneDeltasParser, TrueTypeZoneDeltasCompiler),  # noqa: E501

    # Goes to font again:
    F.fontnames: ("fontnames", NameRecordsParser, NameRecordsCompiler),
    F.CustomCMAPs: ("Custom CMAPs", CustomCmapParser, CustomCmapCompiler),
    F.PCLTTable: ("PCLT Table", PcltParser, PcltCompiler),
    F.ExportPCLTTable: ("Export PCLT Table", Int16Parser, Int16Compiler),
    F.note: ("note", StringParser, StringCompiler),
    F.FontFlags: ("Font Flags", BaseParser, HexStringCompiler),  # FIXME
    F.customdata: ("customdata", StringParser, StringCompiler),
    F.MetricsClassFlags: ("OpenType Metrics Class Flags", OpenTypeMetricsClassFlagsParser, OpenTypeMetricsClassFlagsCompiler),  # noqa: E501
    F.KerningClassFlags: ("OpenType Kerning Class Flags", OpenTypeKerningClassFlagsParser, OpenTypeKerningClassFlagsCompiler),  # noqa: E501

    # Repeat for each binary table:
    # truetypetables: TrueTypeTable
    F.TrueTypeTable: ("TrueTypeTable", BinaryTableParser, BinaryTableCompiler),

    F.features: ("features", OpenTypeStringParser, OpenTypeStringCompiler),

    # Repeat for each OpenType class:
    F.GlyphClass: ("OpenType Class", StringParser, StringCompiler),  # Font.classes

    F.BlockFontInfoEnd: ("Block Font Info End", BaseParser, HexStringCompiler),
    F.BlockMMFontInfoStart: ("Block MM Font Info Start", BaseParser, HexStringCompiler),
    F.AxisCount: ("Axis Count", Int16Parser, Int16Compiler),
    # Repeat for each axis:
    F.AxisName: ("Axis Name", StringParser, StringCompiler),

    F.AnisotropicInterpolationMappings: ("Anisotropic Interpolation Mappings", AnisotropicInterpolationsParser, AnisotropicInterpolationsCompiler),  # noqa: E501
    F.AxisMappingsCount: ("Axis Mappings Count", AxisMappingsCountParser, AxisMappingsCountCompiler),  # noqa: E501
    F.AxisMappings: ("Axis Mappings", AxisMappingsParser, AxisMappingsCompiler),

    # Repeat the next two for each master:
    M.MasterName: ("Master Name", StringParser, StringCompiler),
    M.MasterLocation: ("Master Location", MasterLocationParser, MasterLocationCompiler),

    F.PrimaryInstanceLocations: ("Primary Instance Locations", DoubleListParser, DoubleListCompiler),  # noqa: E501
    F.PrimaryInstances: ("Primary Instances", PrimaryInstancesParser, PrimaryInstancesCompiler),  # noqa: E501

    # Repeat PostScript Info for each master:
    M.PostScriptInfo: ("PostScript Info", PostScriptInfoParser, PostScriptInfoCompiler),

    F.BlockMMFontInfoEnd: ("Block MM Font Info End", BaseParser, HexStringCompiler),
    F.GlobalGuides: ("Global Guides", GlobalGuidesParser, GuidesCompiler),
    F.GlobalGuideProperties: ("Global Guide Properties", GuidePropertiesParser, GuidePropertiesCompiler),  # noqa: E501
    F.GlobalMask: ("Global Mask", GlobalMaskParser, GlobalMaskCompiler),
    F.default_character: ("default_character", StringParser, StringCompiler),

    # Begin: Repeat for each glyph
    G.Glyph: ("Glyph", GlyphParser, GlyphCompiler),
    # Glyph.hlinks and Glyph.vlinks:
    G.Links: ("Links", LinkParser, LinksCompiler),
    G.image: ("image", BackgroundBitmapParser, BackgroundBitmapCompiler),
    G.Bitmaps: ("Glyph Bitmaps", GlyphBitmapsParser, GlyphBitmapsCompiler),
    G.VSB: ("VSB", EncodedValueListParser, EncodedValueListCompiler),  # 1 encoded value per master  # noqa: E501
    G.Sketch: ("Glyph Sketch", GlyphSketchParser, GlyphSketchCompiler),
    G.HintingOptions: ("Glyph Hinting Options", PostScriptGlyphHintingOptionsParser, PostScriptGlyphHintingOptionsCompiler),  # noqa: E501
    G.mask: ("mask", MaskParser, MaskCompiler),
    G.MaskMetrics: ("mask.metrics", MaskMetricsParser, MaskMetricsCompiler),  # Single master mask metrics  # noqa: E501
    G.MaskMetricsMM: ("mask.metrics_mm", MaskMetricsMMParser, MaskMetricsMMCompiler),  # Mask metrics master 2 to 16  # noqa: E501
    G.Origin: ("Glyph Origin", GlyphOriginParser, GlyphOriginCompiler),
    G.unicodes: ("unicodes", GlyphUnicodeParser, GlyphUnicodesCompiler),  # Glyph Unicode  # noqa: E501
    G.CustomDict: ("Custom Dict", StringParser, StringCompiler),
    G.UnicodesNonBMP: ("Glyph Unicode Non-BMP", GlyphUnicodeSuppParser, GlyphUnicodesSuppCompiler),  # noqa: E501
    G.mark: ("mark", Int16Parser, Int16Compiler),  # Mark Color
    G.customdata: ("glyph.customdata", StringParser, StringCompiler),
    G.note: ("glyph.note", StringParser, StringCompiler),
    G.GDEFData: ("Glyph GDEF Data", GlyphGDEFParser, GlyphGDEFCompiler),
    G.AnchorsProperties: ("Glyph Anchors Supplemental", GlyphAnchorsSuppParser, GlyphAnchorsSuppCompiler),  # noqa: E501
    G.AnchorsMM: ("Glyph Anchors MM", GlyphAnchorsParser, GlyphAnchorsCompiler),  # MM-compatible  # noqa: E501
    G.GuideProperties: ("Glyph Guide Properties", GuidePropertiesParser, GuidePropertiesCompiler),  # noqa: E501
    # End: Repeat for each glyph

    F.FontOptions: ("Font Options", FontOptionsParser, FontOptionsCompiler),  # noqa: E501
    F.ExportOptions: ("Export Options", ExportOptionsParser, ExportOptionsCompiler),
    F.MappingMode: ("Mapping Mode", MappingModeParser, MappingModeCompiler),

    # Not seen in FontNames.vfb:
    F.BlockMMKerningStart: ("Block MM Kerning Start", BaseParser, HexStringCompiler),
    F.MMKernPair: ("MMKernPair", MMKernPairParser, MMKernPairCompiler),
    F.BlockMMKerningEnd: ("Block MM Kerning End", BaseParser, HexStringCompiler),

    # File end
    F.BlockFontEnd: ("Block Font End", BaseParser, HexStringCompiler),
    F.BlockFileDataEnd: ("Block File Data End", BaseParser, HexStringCompiler),
}
# fmt: on


# Those entries are ignored in minimal mode:
ignore_minimal = set(
    (
        "Global Guides",
        "Global Mask",
        "Glyph Bitmaps",
        "Glyph Guide Properties",
        "glyph.note",
        "image",
        "mark",
        "mask",
        "note",
    )
)

ignore_minimal_keys = frozenset(
    (
        F.GlobalGuides,
        F.GlobalMask,
        F.note,
        G.Bitmaps,
        G.GuideProperties,
        G.image,
        G.mark,
        G.mask,
        G.note,
    )
)



import logging

from vfbLib.parsers.base import BaseParser
from vfbLib.typing import FLVersionDict

logger = logging.getLogger(__name__)


END = 0
PLATFORM = 1
VERSION = 2
OWNER = 3

PLATFORM_WINDOWS = 0
PLATFORM_MACOS = 1

platform_names = {
    PLATFORM_WINDOWS: "windows",
    PLATFORM_MACOS: "macos",
}

platform_ids = {
    "macos": PLATFORM_MACOS,
    "windows": PLATFORM_WINDOWS,
}


class FLVersionParser(BaseParser):
    """
    A parser that reads the FL Version entry.
    """

    def _parse(self) -> FLVersionDict:
        fl_version_info: FLVersionDict = FLVersionDict(
            platform="macos", version=(5, 2, 2, 128), owner=0
        )
        while True:
            key = self.read_uint8()
            if key == END:
                break
            value = self.read_value()
            if key == PLATFORM:
                fl_version_info["platform"] = platform_names[value]
            elif key == VERSION:
                # Make the version number human-readable
                app_version = []
                for i in (24, 16, 8, 0):
                    # Extract the bytes from the number again
                    app_version.append(value >> i & 0xFF)
                fl_version_info["version"] = tuple(app_version)
            elif key == OWNER:
                fl_version_info["owner"] = value

        return fl_version_info
